import torch

from torch.distributions import Normal

from alfi.models import LFM
from alfi.nn import MuSigmaEncoder
from .operator import NeuralOperator


class NeuralLFM(LFM):
    """
    Implements Neural Process LFM for functions of arbitrary dimensions.

    Parameters
    ----------
    y_dim : int
        Dimension of y values.
    r_dim : int
        Dimension of output representation r.
    z_dim : int
        Dimension of latent variable z.
    h_dim : int
        Dimension of hidden layer in encoder and decoder.
    """

    def __init__(self, block_dim, in_channels, modes, width, r_dim, z_dim):
        super(NeuralLFM, self).__init__()
        self.num_outputs = 1
        self.x_dim = 1
        self.r_dim = r_dim
        self.z_dim = z_dim
        self.block_dim = block_dim
        # Initialize networks
        self.xy_to_r = NeuralOperator(block_dim, in_channels + 1, r_dim, modes, width, num_layers=2)
        self.r_to_mu_sigma = MuSigmaEncoder(r_dim, z_dim)
        self.xz_to_y = NeuralOperator(block_dim, in_channels + z_dim, 2, modes, width, num_layers=2)

    def aggregate(self, r_i):
        """
        Aggregates representations for every (x_i, y_i) pair into a single
        representation.

        Parameters
        ----------
        r_i : torch.Tensor
            Shape (batch_size, s, r_dim)
        """
        dim = list(range(1, r_i.ndim-1))
        return torch.mean(r_i, dim=dim)

    def xy_to_mu_sigma(self, x, y):
        """
        Maps (x, y) pairs into the mu and sigma parameters defining the normal
        distribution of the latent variables z.

        Parameters
        ----------
        x : torch.Tensor
            Shape (batch_size, num_points, x_dim)

        y : torch.Tensor
            Shape (batch_size, num_points, y_dim)
        """
        if self.block_dim == 1:
            batch_size, num_points, _ = x.size()
        else:
            batch_size, num_points, _, _ = x.size()

        # Flatten tensors, as encoder expects one dimensional inputs
        # x_flat = x.view(batch_size * num_points, self.x_dim)
        # y_flat = y.contiguous().view(batch_size * num_points, self.y_dim)
        # Encode each point into a representation r_i
        xy = torch.cat([x, y], dim=-1)
        # print('xy', xy.shape)
        r_i, params_i = self.xy_to_r(xy)

        # Reshape tensors into batches
        # r_i = r_i_flat.view(batch_size, num_points, self.r_dim)
        # Aggregate representations r_i into a single representation r
        r = self.aggregate(r_i)
        # print('agg', r.shape)
        # Return parameters of distribution
        return self.r_to_mu_sigma(r)

    def forward(self, x_context, y_context, x_target, y_target=None):
        """
        Given context pairs (x_context, y_context) and target points x_target,
        returns a distribution over target points y_target.

        Parameters
        ----------
        x_context : torch.Tensor
            Shape (batch_size, num_context, x_dim). Note that x_context is a
            subset of x_target.

        y_context : torch.Tensor
            Shape (batch_size, num_context, y_dim)

        x_target : torch.Tensor
            Shape (batch_size, num_target, x_dim)

        y_target : torch.Tensor or None
            Shape (batch_size, num_target, y_dim). Only used during training.

        Note
        ----
        We follow the convention given in "Empirical Evaluation of Neural
        Process Objectives" where context is a subset of target points. This was
        shown to work best empirically.
        """
        # Infer quantities from tensor dimensions
        if self.training:
            # Encode target and context (context needs to be encoded to
            # calculate kl term)
            mu_target, sigma_target = self.xy_to_mu_sigma(x_target, y_target)
            mu_context, sigma_context = self.xy_to_mu_sigma(x_context, y_context)
            # Sample from encoded distribution using reparameterization trick
            q_target = Normal(mu_target, sigma_target)
            q_context = Normal(mu_context, sigma_context)
            z_sample = q_target.rsample().unsqueeze(1)
            # Get parameters of output distribution
            w1 = x_target.shape[1]
            w2 = x_target.shape[2]
            if self.block_dim == 1:
                xz = torch.cat([x_target, z_sample.repeat(1, w1, 1)], dim=-1)
            else:
                z_sample = z_sample.unsqueeze(1)
                xz = torch.cat([x_target, z_sample.repeat(1, w1, w2, 1)], dim=-1)

            y_pred, y_params = self.xz_to_y(xz)
            sigma = 0.1 + 0.9 * torch.sigmoid(y_pred[..., 1])
            p_y_pred = Normal(y_pred[..., 0], sigma)
            return p_y_pred, y_params, q_target, q_context
        else:
            # At testing time, encode only context
            mu_context, sigma_context = self.xy_to_mu_sigma(x_context, y_context)
            # Sample from distribution based on context
            q_context = Normal(mu_context, sigma_context)
            z_sample = q_context.rsample().unsqueeze(1)
            # Predict target points based on context
            w1 = x_target.shape[1]
            w2 = x_target.shape[2]
            if self.block_dim == 1:
                xz = torch.cat([x_target, z_sample.repeat(1, w1, 1)], dim=-1)
            else:
                z_sample = z_sample.unsqueeze(1)
                xz = torch.cat([x_target, z_sample.repeat(1, w1, w2, 1)], dim=-1)

            y_pred, y_params = self.xz_to_y(xz)
            sigma = 0.1 + 0.9 * torch.sigmoid(y_pred[..., 1])
            p_y_pred = Normal(y_pred[..., 0], sigma)
            return p_y_pred, y_params

    def count_params(self):
        return -1



from typing import List, Callable
import torch

from alfi.trainers import Trainer
from alfi.models import VariationalLFM, TrainMode
try:
    from alfi.models import PartialLFM
except:
    class PartialLFM:
        pass
from alfi.utilities.torch import spline_interpolate_gradient
import numpy as np

class PreEstimator(Trainer):
    def __init__(self,
                 model: VariationalLFM,
                 optimizers: List[torch.optim.Optimizer],
                 dataset,
                 **kwargs):
        super().__init__(model, optimizers, dataset, **kwargs)
        data = next(iter(self.data_loader))
        t, y = data[0][0], data[1]

        t_interpolate, self.y_interpolate, self.target = self.interpolate(t, y)
        self.input_pair = (t_interpolate, self.y_interpolate)
        self.model_kwargs = {}

    def interpolate(self, time, y, num_intermediate=9):
        """
        :param time: pseudotime
        :param y: output
        :param num_intermediate: this determines how granular the interpolation is
        :return:
        """
        print('interpolating', y.shape)
        # y[y <= 0] = np.nan
        if y.ndim > 2:
            ts = list()
            ys = list()
            t_interp = torch.linspace(time.min(), time.max(), 200)
            y_interpolate = np.zeros((y.shape[0], y.shape[1], 200))
            y_grad = torch.zeros((y.shape[0], y.shape[1], 200))
            t_interpolate = t_interp#.unsqueeze(0).repeat((y.shape[0], 1))
            print(t_interpolate.shape)
            time_uniq, ind = np.unique(time, return_index=True)
            print(time)
            for i in range(y.shape[0]):
                _t_interpolate, _y_interpolate, _y_grad, _ = spline_interpolate_gradient(
                    time[ind], y[i:i+1, ..., ind].transpose(-1, -2), num_intermediate, x_interpolate=t_interp)
                y_interpolate[i, :] = _y_interpolate.transpose(-1, -2)
                y_grad[i, :] = _y_grad.transpose(-1, -2)

        else:
            t_interpolate, y_interpolate, y_grad, _ = spline_interpolate_gradient(
                time, y.transpose(-1, -2), num_intermediate)

            y_interpolate = y_interpolate.transpose(-1, -2)
        return t_interpolate, y_interpolate, y_grad

    def single_epoch(self, epoch=0, **kwargs):
        assert self.model.train_mode == TrainMode.GRADIENT_MATCH
        [optim.zero_grad() for optim in self.optimizers]
        # y = y.cuda() if is_cuda() else y
        output = self.model(self.input_pair, **self.model_kwargs)
        y_target = self.target

        log_likelihood, kl_divergence, _ = self.model.loss_fn(output, y_target, mask=self.train_mask)

        loss = - (log_likelihood - kl_divergence)
        loss.backward()
        [optim.step() for optim in self.optimizers]
        self.debug_out(output, y_target)
        return loss.item(), (-log_likelihood.item(), kl_divergence.item())

    def debug_out(self, output, y_target):
        pass


class PartialPreEstimator(PreEstimator):
    def __init__(self,
                 model: PartialLFM,
                 optimizers: List[torch.optim.Optimizer],
                 dataset,
                 pde_func: Callable,
                 input_pair,
                 target,
                 **kwargs):
        super(PreEstimator, self).__init__(model, optimizers, dataset, **kwargs)

        self.pde_func = pde_func
        self.input_pair = input_pair
        self.target = target
        disc = dataset.disc if hasattr(dataset, 'disc') else 1

        self.model_kwargs = dict(step=disc, pde_func=self.pde_func)

    def debug_out(self, output, y_target):
        # plt.figure()
        # plt.imshow(output.mean.detach().view(41, 41).t())
        # plt.colorbar()
        pass


import torch
import math
import numpy as np
from typing import Optional
from torchcubicspline import(natural_cubic_spline_coeffs,
                             NaturalCubicSpline)
from scipy.interpolate import UnivariateSpline
from scipy.signal import savgol_filter

CUDA_AVAILABLE = False


def is_cuda():
    import torch
    return CUDA_AVAILABLE and torch.cuda.is_available()


def discretisation_length(N, d):
    """Returns the length of a linspace where there are N points with d intermediate (in-between) points."""
    return (N - 1) * (d + 1) + 1


def spline_interpolate_gradient(x: torch.Tensor, y: torch.Tensor, num_disc=9, x_interpolate: Optional[torch.Tensor] = None):
    """
    y should be of shape (..., length, input_channels)
    Returns x_interpolate, y_interpolate, y_grad, y_grad_2: the interpolated time, data and gradient
    """
    if x_interpolate is None:
        x_interpolate = torch.linspace(x.min(), x.max(), discretisation_length(x.shape[0], num_disc), device=x.device)
    coeffs = natural_cubic_spline_coeffs(x, y)
    spline = NaturalCubicSpline(coeffs)
    y_interpolate = spline.evaluate(x_interpolate)
    y_grad = spline.derivative(x_interpolate) #y_interpolate, denom, axis=1)
    y_grad_2 = spline.derivative(x_interpolate, order=2)
    return x_interpolate, y_interpolate, y_grad, y_grad_2


def savgol_filter_gradient(x, y, window_length=75, polyorder=3):
    """
    Applies the Savitzky-Golay filter to denoise the series y, and returns the first derivative
    """
    savgol_config = dict(
        window_length=window_length,
        polyorder=polyorder
    )
    dx = x[1] - x[0]
    y_denoised = savgol_filter(y.squeeze(), **savgol_config)
    dy = savgol_filter(y.squeeze(), **savgol_config, deriv=1, delta=dx)
    return y_denoised, dy


"""These metrics are translated from https://rdrr.io/cran/lineqGPR/man/errorMeasureRegress.html"""
def smse(y_test, f_mean):
    """Standardised mean square error (standardised by variance)"""
    return (y_test - f_mean).square() / y_test.var()


def q2(y_test, f_mean):
    y_mean = y_test.mean()
    return 1 - (y_test - f_mean).square().sum() / (y_test - y_mean).square().sum()


def cia(y_test, f_mean, f_var, n=1):
    return ((y_test >= (f_mean - n * f_var.sqrt())) &
            (y_test <= (f_mean + n * f_var.sqrt()))).double().mean()


def ceil(x):
    return int(math.ceil(x))


def softplus(value):
    return torch.log(1 + torch.exp(value))


def inv_softplus(value):
    return torch.log(torch.exp(value) - 1)


def cholesky_inverse(cholesky_factor, upper=False):
    """Courtesy of Alex Campbell"""
    batch_shape = list(cholesky_factor.shape[:-2])
    matrix_dim = list(cholesky_factor.shape[-2:])
    if batch_shape:
        flat_batch = torch.tensor(batch_shape).prod()
        inv_cholesky_factor = torch.stack([
            torch.cholesky_inverse(
                cholesky_factor.view(-1, *matrix_dim)[l], upper=upper)
            for l in range(flat_batch)])
        matrix = inv_cholesky_factor.view(*batch_shape, *matrix_dim)
    else:
        matrix = torch.cholesky_inverse(cholesky_factor, upper=upper)
    return matrix


def get_image(data, intensity_index=2):
    """
    Returns an array compatible with plt.imshow
    Parameters:
        data: should be of shape (N, D) where N is the number of datapoints and D is the number of columns.
              First two columns are temporal and spatial dimensions
        intensity_index: the column index of the intensity of the image
    """
    ts = np.unique(data[:, 0])
    rows = list()
    for t in ts:
        row = data[data[:, 0] == t, intensity_index]
        rows.extend([row])
    return np.array(rows).T

def discretise(time, num_discretised=40):
    """
    Calculates a discretised time grid with num_discretised points.
    Note: the return size is of size num_discretised + 1
    @param time: the time vector of size (2, N) where dim 0 is time, dim 1 is space.
    @param num_discretised: the number of points in the time grid
    @return: grid, time: discretised time grid as well as an updated time vector which is
    down or upsampled to include all points in the discretised time grid
    """
    t = np.unique(time)
    t.sort()
    t_range = t[-1] - t[0]
    dp = t_range / num_discretised
    return np.arange(t[0], t[-1] + dp, dp)


def get_mean_trace(trace):
    mean_trace = dict()
    for key in trace.keys():
        params = torch.stack(trace[key])
        for i in range(1, params.ndim):
            params = params.mean(-1)
        mean_trace[key] = params
    return mean_trace

def compl_mul1d(a, b):
    # (batch, in_channel, x ), (in_channel, out_channel, x) -> (batch, out_channel, x)
    return torch.einsum('bix,iox->box', a, b)

def compl_mul2d(a, b):
    # (batch, in_channel, x,y ), (in_channel, out_channel, x,y) -> (batch, out_channel, x,y)
    return torch.einsum('bixy,ioxy->boxy', a, b)



import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
from mpl_toolkits.axes_grid1 import ImageGrid
import numpy as np
import seaborn as sns
import pandas as pd
from seaborn import kdeplot
from .colours import Colours

plt.style.use('seaborn')
plt.rcParams['font.family'] = 'serif'
plt.rcParams['font.serif'] = 'CMU Serif'
sns.set(style='white', font="CMU Serif")


def plot_spatiotemporal_data(images, extent, nrows=1, ncols=None, titles=None, figsize=(10, 5), cticks=None, clim=None):
    if ncols == None:
        ncols = len(images)
    fig = plt.figure(figsize=figsize)
    grid = ImageGrid(fig, 111,  # similar to subplot(144)
                     nrows_ncols=(nrows, ncols),
                     axes_pad=(0.8, 0.1),
                     label_mode='all',
                     share_all=False,
                     cbar_location="right",
                     cbar_mode="each",
                     cbar_size="7%",
                     cbar_pad="2%",
                 )
    aspect = (extent[1]-extent[0]) / (extent[3]-extent[2])
    titles = [None] * len(images) if titles is None else titles
    plotnum = 0
    for ax, cax, image, title in zip(grid, grid.cbar_axes, images, titles):
        im = ax.imshow(image, extent=extent, origin='lower', aspect=aspect)
        cb = plt.colorbar(im, cax=cax)
        if cticks is None:
            cb.ax.yaxis.set_major_locator(ticker.MaxNLocator(nbins=5, steps=[1, 2, 4, 5, 10], integer=True))
        else:
            cb.set_ticks(cticks)
        if clim is not None:
            im.set_clim(clim[plotnum])
        ax.set_xticks([np.ceil(extent[0]), np.floor(extent[1])])
        ax.set_yticks([np.ceil(extent[2]), np.floor(extent[3])])
        ax.set_xlim([extent[0], extent[1]])
        ax.set_ylim([extent[2], extent[3]])
        if (plotnum % ncols) == 0:
            ax.set_ylabel('x')
        ax.set_xlabel('t')
        if title is not None:
            ax.set_title(title)
        plotnum += 1
    return grid


def plot_phase(x_samples, y_samples,
               x_target=None, y_target=None,
               x_mean=None, y_mean=None, figsize=(4, 4), ax=None, pred_label='Prediction'):
    """

    @param x_samples:
    @param y_samples:
    @param x_mean: if None, estimated from data
    @param y_mean: if None, estimated from data
    @return:
    """
    if x_mean is None:
        x_mean = x_samples.mean(0)
    if y_mean is None:
        y_mean = y_samples.mean(0)
    if ax is None:
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(1, 1, 1)
    ax.plot(x_mean, y_mean, label=pred_label, color=Colours.line_color)
    if x_target is not None:
        ax.plot(x_target, y_target, label='Target', color=Colours.scatter_color)

    data_stacked = np.stack([x_samples.flatten(), y_samples.flatten()])

    ndp_df = pd.DataFrame(data_stacked.transpose(), columns=['Prey', 'Predator'])

    kdeplot(data=ndp_df, ax=ax, fill=True, x="Prey", y="Predator",
            color='pink', alpha=0.1, levels=3, thresh=.1, )
    plt.legend(loc='upper right')


def plot_variational_dist(lfm):
    from gpytorch.variational import CholeskyVariationalDistribution
    plt.figure()
    strat = lfm.gp_model.variational_strategy.base_variational_strategy
    dist = strat._variational_distribution
    if type(dist) is CholeskyVariationalDistribution:
        mean = dist.variational_mean.detach()
        covar = dist.chol_variational_covar.detach()
    else:
        mean = dist.natural_vec.detach()
        covar = dist.natural_mat.detach()
    xy = strat.inducing_points.squeeze()
    plt.scatter(xy[:, 0], xy[:, 1], c=mean)
    plt.colorbar()
    plt.figure()
    plt.imshow(covar.squeeze())
    plt.colorbar()



from alfi.datasets import DataHolder
from alfi.configuration import MCMCConfiguration
import abc

class MCMCLFM():

    def __init__(self, data: DataHolder, options: MCMCConfiguration):
        self.options = options
        self.data = data
        self.preprocessing_variance = options.preprocessing_variance
        self.num_outputs = data.m_obs.shape[1]
        self.num_tfs = data.f_obs.shape[1]
        self.num_replicates = data.f_obs.shape[0]
        self.parameter_state = None

    @abc.abstractmethod
    def sample(self, T=2000, **kwargs):
        pass



from typing import List

import torch

from alfi.utilities.torch import is_cuda

from alfi.trainers import Trainer
from alfi.models import VariationalLFM


class VariationalTrainer(Trainer):
    """
    Parameters:
        batch_size: in the case of the transcriptional regulation model, we train the entire gene set as a batch
    """
    def __init__(self,
                 lfm: VariationalLFM,
                 optimizers: List[torch.optim.Optimizer],
                 dataset,
                 warm_variational=-1,
                 **kwargs):
        super().__init__(lfm, optimizers, dataset, batch_size=lfm.num_tasks, **kwargs)
        self.warm_variational = warm_variational
        # if warm_variational >= 0:  # Cold start: don't train non variational parameters initially.
        #     for param in self.model.nonvariational_parameters():
        #         param.requires_grad = False

    def single_epoch(self, step_size=1e-1, epoch=0, **kwargs):
        epoch_loss = 0
        epoch_ll = 0
        epoch_kl = 0
        for i, data in enumerate(self.data_loader):
            [optim.zero_grad() for optim in self.optimizers]
            data_input, y = data
            data_input = data_input.cuda() if is_cuda() else data_input
            y = y.cuda() if is_cuda() else y
            # Assume that the batch of t s are the same
            data_input, y = data_input[0], y
            output = self.model(data_input, step_size=step_size)
            y_target = y.t()

            self.debug_out(data_input, y, output)
            log_likelihood, kl_divergence, _ = self.model.loss_fn(output, y_target, mask=self.train_mask)

            loss = - (log_likelihood - kl_divergence)

            loss.backward()
            if epoch >= self.warm_variational:
                [optim.step() for optim in self.optimizers]
            else:
                self.optimizers[0].step()

            epoch_loss += loss.item()
            epoch_ll += log_likelihood.item()
            epoch_kl += kl_divergence.item()
            # if (epoch % 10) == 0:
            #     print(dict(self.model.gp_model.named_variational_parameters()))
            # Now we are warmed up, start training non variational parameters in the next epoch.
            # if epoch + 1 == self.warm_variational:
            #     for param in self.model.nonvariational_parameters():
            #         param.requires_grad = True

        return epoch_loss, (-epoch_ll, epoch_kl)

    def debug_out(self, data_input, y_target, output):
        pass



class ParamGroupMixin:
    """
    This mixin adds the param_group property which is the ordered list of parameters
    associated with a sampler. 
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._param_group = list()

    @property
    def param_group(self):
        return self._param_group

    @param_group.setter
    def param_group(self, value):
        self._param_group = value

    def __str__(self):
        param_names = [param.name for param in self.param_group]
        return '{} {}'.format(self.__class__.__name__, param_names)



import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from alfi.datasets import scaled_barenco_data
from alfi.models import VariationalLFM
from .colours import Colours
from .base_plotter import Plotter


class Plotter1d(Plotter):
    """
    This Plotter is designed for gp models.
    """

    def __init__(self, model, output_names, style='seaborn'):
        super().__init__(model, output_names, style=style)
        self.num_outputs = self.output_names.shape[0]
        self.num_replicates = self.model.num_outputs // self.num_outputs
        self.variational = isinstance(self.model, VariationalLFM)

    def plot_barenco(self, mean):
        barenco_f, _ = scaled_barenco_data(mean)
        plt.scatter(np.linspace(0, 12, 7), barenco_f, marker='x', s=60, linewidth=2, label='Barenco et al.')

    def plot_gp(self, gp,
                t_predict, t_scatter=None, y_scatter=None,
                num_samples=0,
                transform=lambda x:x,
                ylim=None,
                titles=None,
                max_plots=10,
                replicate=0,
                only_plot_index=None,
                ax=None,
                plot_inducing=False,
                color=Colours.line_color,
                shade_color=Colours.shade_color):
        """
        Parameters:
            gp: output distribution of LFM or associated GP models.
            t_predict: tensor (T*,) prediction input vector
            t_scatter: tensor (T,) target input vector
            y_scatter: tensor (J, T) target output vector
        """
        mean = gp.mean.detach().transpose(0, 1)  # (T, J)
        std = gp.variance.detach().transpose(0, 1).sqrt()
        num_plots = mean.shape[0]
        mean = mean.view(num_plots, self.num_replicates, -1).transpose(0, 1)[replicate].detach()
        std = std.view(num_plots, self.num_replicates, -1).transpose(0, 1)[replicate].detach()
        mean = transform(mean)
        # std = transform(std)
        num_plots = min(max_plots, num_plots)
        axes_given = ax is not None
        if not axes_given:
            fig = plt.figure(figsize=(6, 4 * np.ceil(num_plots / 3)))
        for i in range(num_plots):
            if only_plot_index is not None:
                i = only_plot_index
            if not axes_given:
                ax = fig.add_subplot(num_plots, min(num_plots, 3), i + 1)
            if titles is not None:
                ax.set_title(titles[i])

            ax.plot(t_predict, mean[i], color=color)
            ax.fill_between(t_predict,
                            mean[i] + 2 * std[i],
                            mean[i] - 2 * std[i],
                            color=shade_color, alpha=0.3)

            for _ in range(num_samples):
                ax.plot(t_predict, transform(gp.sample().detach()).transpose(0, 1)[i], alpha=0.3, color=color)

            if self.variational and plot_inducing:
                inducing_points = self.model.inducing_points.detach()[0].squeeze()
                ax.scatter(inducing_points, np.zeros_like(inducing_points), marker='_', c='black', linewidths=2)
            if t_scatter is not None:
                ax.scatter(t_scatter, y_scatter[replicate, i], color=Colours.scatter_color, marker='x')
            if ylim is None:
                lb = min(mean[i])
                lb -= 0.2 * lb
                ub = max(mean[i]) * 1.2
                ax.set_ylim(lb, ub)
            else:
                ax.set_ylim(ylim)

            if axes_given:
                break
        return gp


    def plot_convergence(self, trainer):
        titles = ['basal', 'decay', 'sensitivity', 'lengthscale']
        datas = [np.array(trainer.basalrates)[:,:,0],
                 np.array(trainer.decayrates)[:,:,0],
                 np.array(trainer.sensitivities)[:,:,0],
                 np.array(trainer.lengthscales)[:, 0, 0]]

        plt.figure(figsize=(5, 6))
        for i, (title, data) in enumerate(zip(titles, datas)):
            plt.subplot(411 + i)
            plt.title(title)
            # if data.ndim > 1:
            #     for j in range(data.shape[1]):

            plt.plot(data)



import torch

from torch.nn import Module, Linear


class MuSigmaEncoder(Module):
    """
    Maps a representation r to mu and sigma which will define the normal
    distribution from which we sample the latent variable z.

    Parameters
    ----------
    r_dim : int
        Dimension of output representation r.

    z_dim : int
        Dimension of latent variable z.
    """

    def __init__(self, r_dim, z_dim):
        super(MuSigmaEncoder, self).__init__()

        self.r_dim = r_dim
        self.z_dim = z_dim

        self.r_to_hidden = Linear(r_dim, r_dim)
        self.hidden_to_mu = Linear(r_dim, z_dim)
        self.hidden_to_sigma = Linear(r_dim, z_dim)

    def forward(self, r):
        """
        r : torch.Tensor
            Shape (batch_size, r_dim)
        """
        hidden = torch.relu(self.r_to_hidden(r))
        mu = self.hidden_to_mu(hidden)
        # Define sigma following convention in "Empirical Evaluation of Neural
        # Process Objectives" and "Attentive Neural Processes"
        sigma = 0.1 + 0.9 * torch.sigmoid(self.hidden_to_sigma(hidden))
        return mu, sigma



from typing import List

import torch
import numpy as np
from matplotlib import pyplot as plt
from datetime import datetime

from alfi.utilities.torch import is_cuda, discretise
from alfi.models import PartialLFM
from alfi.plot import plot_spatiotemporal_data
from .variational import VariationalTrainer
from alfi.utilities.torch import cia, q2, smse, softplus


class PDETrainer(VariationalTrainer):

    def __init__(self, lfm: PartialLFM, optimizers: List[torch.optim.Optimizer], dataset, clamp=False, lf_target=None, **kwargs):
        super().__init__(lfm, optimizers, dataset, **kwargs)
        self.debug_iteration = 0
        self.clamp = clamp
        self.plot_outputs = True
        self.plot_outputs_iter = 10
        self.prot_q2_best = 0
        self.mrna_q2_best = 0
        self.cia = (0, 0)
        self.disc = 1
        data = next(iter(dataset))
        data_input, y = data
        tx, y_target = data_input, y

        # 1. Ensure that time axis must be in ascending order:
        # We use mergesort to maintain relative order
        self.t_sorted = np.argsort(tx[0, :], kind='mergesort')
        tx = tx[:, self.t_sorted]
        self.num_t_orig = tx[0].unique().shape[0]
        self.num_x_orig = tx[1].unique().shape[0]
        y_target = y_target[:, self.t_sorted]

        # 2. Discretise time
        if hasattr(dataset, 'num_discretised'):
            num_discretised = dataset.num_discretised
            self.disc = dataset.disc
        else:
            num_discretised = 40
        time = discretise(tx[0, :], num_discretised=num_discretised)
        time = torch.tensor(time)

        # 3. Discretise space
        spatial_grid = self.discretise_spatial(tx)
        spatial = torch.tensor(spatial_grid)

        # 4. Reconstruct dataset
        new_t = time.repeat(spatial.shape[0], 1).transpose(0, 1).reshape(-1)
        # t_mask = new_t == tx[0, :]

        new_x = spatial.repeat(time.shape[0])
        # x_mask = new_x == tx[1, :]

        # mask = torch.stack([t_mask, x_mask])
        self.tx = torch.stack([new_t, new_x])
        self.y_target = y_target.cuda() if is_cuda() else y_target
        self.tx = self.tx.cuda() if is_cuda() else self.tx
        self.t_sorted = self.t_sorted.cuda() if is_cuda() else self.t_sorted
        self.lf_target = lf_target[self.t_sorted, 2]
        self.num_t = self.tx[0, :].unique().shape[0]
        self.num_x = self.tx[1, :].unique().shape[0]

    def discretise_spatial(self, tx):
        # whilst maintaining an inverse mapping to mask the output for
        # calculating the loss. shape (T, X_unique).
        # For each time, for which spatial points do we have data for
        spatial = torch.unique(tx[1, :])
        range = spatial[-1] - spatial[0]
        x_dp = spatial[1] - spatial[0]
        print('x dp is set to', x_dp)
        num_discretised = int(range / x_dp)
        spatial_grid = discretise(spatial, num_discretised=num_discretised)
        return spatial_grid

    def single_epoch(self, step_size=1e-1, epoch=0, **kwargs):
        [optim.zero_grad() for optim in self.optimizers]
        output = self.model(self.tx, step_size=step_size, step=self.disc)
        y_target = self.y_target.t()

        self.debug_out(self.tx, y_target, output)
        log_likelihood, kl_divergence, _ = self.model.loss_fn(output, y_target, mask=self.train_mask)
        loss = - (log_likelihood - kl_divergence)

        loss.backward()
        if epoch >= self.warm_variational:
            [optim.step() for optim in self.optimizers]
        else:
            print(epoch, 'warming up')
            self.optimizers[0].step()

        # Now we are warmed up, start training non variational parameters in the next epoch.
        if epoch + 1 == self.warm_variational:
            for param in self.model.nonvariational_parameters():
                param.requires_grad = True

        return loss.item(), (-log_likelihood.item(), kl_divergence.item())

    def debug_out(self, data_input, y_target, output):
        if (self.debug_iteration % 1) != 0:
            self.debug_iteration += 1
            return
        print('Mean output variance:', output.variance.mean().item())
        if self.train_mask is not None:
            with torch.no_grad():
                log_likelihood, kl_divergence, _ = self.model.loss_fn(output, y_target, mask=~self.train_mask)
                test_loss = - (log_likelihood - kl_divergence)
            print('Test loss:', test_loss.item())
        with torch.no_grad():
            f_mean = output.mean
            f_var = output.variance
            prot_q2 = q2(y_target.squeeze(), f_mean.squeeze())
            prot_cia = cia(y_target.squeeze(), f_mean.squeeze(), f_var.squeeze())
            prot_mse = torch.mean(torch.square(y_target.squeeze() - f_mean.squeeze()))
            gp = self.model.gp_model(self.tx.t())
            lf_target = self.lf_target
            num_t = self.num_t
            num_x = self.num_x
            f_mean = gp.mean.view(num_t, num_x)[::self.disc].reshape(-1)
            f_var = gp.variance.view(num_t, num_x)[::self.disc].reshape(-1)

            mrna_q2 = q2(lf_target.squeeze(), f_mean.squeeze())
            mrna_cia = cia(lf_target.squeeze(), f_mean.squeeze(), f_var.squeeze())
            mrna_mse = torch.mean(torch.square(lf_target.squeeze() - f_mean.squeeze()))

            if (mrna_q2 + prot_q2) > (self.mrna_q2_best + self.prot_q2_best):
                self.mrna_q2_best = mrna_q2
                self.prot_q2_best = prot_q2
                self.cia = (mrna_cia, prot_cia)
            print(f'prot mse: {prot_mse.item():.05f}')
            print(f'mrna mse: {mrna_mse.item():.05f}')
            print(f'prot Q2: {prot_q2.item():.03f}')
            print(f'prot CA: {prot_cia.item():.03f}')
            print(f'mrna Q2: {mrna_q2.item():.03f}')
            print(f'mrna CA: {mrna_cia.item():.03f}')

        if self.plot_outputs and (self.debug_iteration % self.plot_outputs_iter) == 0:
            ts = self.tx[0, :].unique().numpy()
            xs = self.tx[1, :].unique().numpy()
            extent = [ts[0], ts[-1], xs[0], xs[-1]]

            num_t = self.num_t_orig
            num_x = self.num_x_orig
            f_mean = output.mean.reshape(num_t, num_x).detach()
            y_target = y_target.reshape(num_t, num_x)
            axes = plot_spatiotemporal_data(
                [f_mean.t(), y_target.t()],
                extent, titles=['Prediction', 'Ground truth']
            )
            xy = self.model.inducing_points.detach()[0]
            axes[0].scatter(xy[:, 0], xy[:, 1], facecolors='none', edgecolors='r', s=3)

            plt.savefig(str(datetime.now().timestamp()) + '.png')
            self.model.save('currentmodel')
        self.debug_iteration += 1

    def print_extra(self):
        print(' s:', softplus(self.model.fenics_parameters[0][0]).item(),
              'dec:', softplus(self.model.fenics_parameters[1][0]).item(),
              'diff:', softplus(self.model.fenics_parameters[2][0]).item())
        super().print_extra()

    def after_epoch(self):
        super().after_epoch()
        if self.clamp:
            with torch.no_grad():
                self.model.fenics_parameters[2].clamp_(-15, -2.25)
                self.model.fenics_parameters[1].clamp_(-15, -2.25)
                self.model.fenics_parameters[0].clamp_(-15, -2.25)



from fenics import *
from fenics_adjoint import *
from dolfin import *

import numpy as np


def interval_mesh(spatial):
    domain_vertices = [Point(x) for x in spatial]
    editor = MeshEditor()
    mesh = Mesh()
    editor.open(mesh, 'interval', 1, 1)
    editor.init_vertices(len(domain_vertices))
    editor.init_cells(len(domain_vertices) - 1)

    for i, vertex in enumerate(domain_vertices):
        editor.add_vertex(i, vertex)
        if i < len(domain_vertices) - 1:
            editor.add_cell(i, np.array([i, i + 1]))

    return mesh



import torch
from pathlib import Path
from torch.nn import Parameter
from gpytorch.likelihoods import MultitaskGaussianLikelihood
from gpytorch.mlls import VariationalELBO
import numpy as np
from tqdm import tqdm

from matplotlib import pyplot as plt

from .datasets import TranscriptomicTimeSeries, LFMDataset
from alfi.configuration import VariationalConfiguration
from alfi.utilities.torch import softplus, discretisation_length


class ToyTranscriptomics(LFMDataset):
    def __init__(self, data_dir='../data'):
        data = torch.load(Path(data_dir) / 'toy_transcriptomics.pt')
        x_train = data['x_train']
        y_train = data['y_train']
        x_test = data['x_test']
        y_test = data['y_test']
        params_train = data['params_train']
        params_test = data['params_test']
        ntrain = x_train.shape[0]
        ntest = x_test.shape[0]

        train = [(x_train[i].type(torch.float),
                  y_train[i].type(torch.float),
                  params_train[i].type(torch.float)) for i in range(ntrain)]
        test = [(x_test[i].type(torch.float),
                 y_test[i].type(torch.float),
                 params_test[i].type(torch.float)) for i in range(ntest)]
        self._data = train
        self.train_data = train
        self.test_data = test


class ToyTranscriptomicGenerator(LFMDataset):

    def __init__(self,
                 num_outputs=30,
                 num_latents=3,
                 num_times=10,
                 plot=False,
                 softplus=True,
                 latent_data_present=False,
                 dtype=torch.float32):
        self.num_outputs = num_outputs
        self.num_latents = num_latents
        self.num_times = num_times
        self.num_disc = 9
        self.plot = plot
        self.gene_names = np.arange(num_outputs)
        self.softplus = softplus
        self.dtype = dtype
        self.latent_data_present = latent_data_present
    def pick_decay(self):
        return 0.2 + .5 * torch.rand(torch.Size([self.num_outputs, 1]), dtype=self.dtype)

    def pick_sensitivity(self):
        return 0.3 + 2 * torch.rand(torch.Size([self.num_outputs, 1]), dtype=self.dtype)

    def pick_basal(self):
        return 0.01 + 0.2 * torch.rand(torch.Size([self.num_outputs, 1]), dtype=self.dtype)

    def generate(self, ntrain, ntest, data_dir):
        datasets = list()
        params = list()
        for _ in tqdm(range(ntrain + ntest)):
            basal_rate = self.pick_basal()
            sensitivity = self.pick_sensitivity()
            decay_rate = self.pick_decay()
            lengthscale = 0.1 + 3 * torch.rand(torch.Size([1]), dtype=self.dtype)[0]
            params.append(torch.stack([basal_rate, sensitivity, decay_rate]))
            dataset = self.generate_single(basal_rate, sensitivity, decay_rate, lengthscale=lengthscale)
            datasets.append([dataset.m_observed, dataset.f_observed])
        x_train = torch.cat([obs[0] for obs in datasets[:ntrain]]).permute(0, 2, 1)
        x_test = torch.cat([obs[0] for obs in datasets[ntrain:]]).permute(0, 2, 1)
        grid = self.t_observed.reshape(1, -1, 1).repeat(ntrain, 1, 1)  # (1, 32, 32, 40, 1)
        grid_test = self.t_observed.reshape(1, -1, 1).repeat(ntest, 1, 1)  # (1, 32, 32, 40, 1)

        params = torch.stack(params, dim=0)
        print(params.shape)
        params_train = params[:ntrain].squeeze(-1)
        params_test = params[ntrain:].squeeze(-1)

        x_train = torch.cat([grid, x_train], dim=-1)
        x_test = torch.cat([grid_test, x_test], dim=-1)
        y_train = torch.cat([obs[1] for obs in datasets[:ntrain]]).permute(0, 2, 1)
        y_test = torch.cat([obs[1] for obs in datasets[ntrain:]]).permute(0, 2, 1)

        torch.save(dict(x_train=x_train, x_test=x_test,
                        y_train=y_train, y_test=y_test,
                        params_train=params_train, params_test=params_test), Path(data_dir) / 'toy_transcriptomics.pt')

    def generate_single(self, basal=None, sensitivity=None, decay=None, lengthscale=2., epochs=150):
        from alfi.models import OrdinaryLFM, generate_multioutput_gp
        self.epochs = epochs
        self.decay = decay if decay is not None else self.pick_decay()
        self.basal = basal if basal is not None else self.pick_basal()
        self.sensitivity = sensitivity if sensitivity is not None else self.pick_sensitivity()
        self.lengthscale = lengthscale

        class ToyLFM(OrdinaryLFM):
            """
            This LFM is to generate toy data.
            """

            def __init__(self, num_outputs, gp_model, ref, config: VariationalConfiguration, **kwargs):
                super().__init__(num_outputs, gp_model, config, **kwargs)
                self.ref = ref
                self.dtype = ref.dtype
                num_latents = gp_model.variational_strategy.num_tasks
                basal = torch.tensor(ref.basal, dtype=self.dtype)
                decay = torch.tensor(ref.decay, dtype=self.dtype)
                sensitivity = torch.tensor(ref.sensitivity, dtype=self.dtype)
                self.decay_rate = Parameter(decay)
                self.basal_rate = Parameter(basal)
                self.sensitivity = Parameter(sensitivity)
                weight = 0.5 + 1 * torch.randn(torch.Size([self.num_outputs, num_latents]), dtype=self.dtype)
                num_remove = num_outputs // 2
                weight[torch.randperm(self.num_outputs)[:num_remove], torch.randint(num_latents, [num_remove])] = 0
                self.weight = Parameter(weight)
                self.weight_bias = Parameter(torch.randn(torch.Size([self.num_outputs, 1]), dtype=self.dtype))

            def initial_state(self):
                return self.basal_rate / self.decay_rate

            def odefunc(self, t, h):
                """h is of shape (num_samples, num_outputs, 1)"""
                self.nfe += 1
                decay = self.decay_rate * h
                f = self.latent_gp[:, :, self.t_index].unsqueeze(2)
                h = self.basal_rate + f - decay
                if t > self.last_t:
                    self.t_index += 1
                self.last_t = t
                return h

            def nonlinearity(self, f):
                if self.ref.softplus:
                    return softplus(f)
                return f

            def mix(self, f):
                if f.shape[1] > 1:
                    eps = torch.tensor(torch.finfo(self.dtype).tiny, dtype=self.dtype)
                    interactions = torch.matmul(self.weight, torch.log(f + eps)) + self.weight_bias
                    f = torch.sigmoid(interactions)  # TF Activation Function (sigmoid)
                else:
                    f = f.repeat(1, self.num_outputs, 1)
                return f

        config = VariationalConfiguration(
            num_samples=70,
            initial_conditions=False # TODO
        )

        num_inducing = 10  # (I x m x 1)
        inducing_points = torch.linspace(0, self.num_times-1, num_inducing, dtype=self.dtype).repeat(self.num_latents, 1).view(self.num_latents, num_inducing, 1)
        t_predict = torch.linspace(0, self.num_times-1, discretisation_length(self.num_times, self.num_disc), dtype=self.dtype)

        gp_model = generate_multioutput_gp(self.num_latents, inducing_points,
                                           initial_lengthscale=lengthscale,
                                           gp_kwargs=dict(natural=False))
        self.train_gp(gp_model, t_predict)
        with torch.no_grad():
            self.lfm = ToyLFM(self.num_outputs, gp_model, self, config)
            q_m = self.lfm.predict_m(t_predict)
            self.t_observed_highres = t_predict
            self.t_observed = t_predict[::self.num_disc+1]
            self.m_observed_highres = q_m.mean.unsqueeze(0).permute(0, 2, 1)
            self.m_observed = q_m.mean[::self.num_disc+1].unsqueeze(0).permute(0, 2, 1)
            self.data = [(self.t_observed, self.m_observed[0, i]) for i in range(self.num_outputs)]
            if self.latent_data_present:
                self.data.extend([(self.t_observed, self.f_observed[0, i]) for i in range(self.num_latents)])
        return self

    def train_gp(self, gp_model, t_predict):
        q_f = gp_model(t_predict)
        samples = q_f.sample(torch.Size([1]))  # sample from the prior
        if self.plot:
            fig, axes = plt.subplots(ncols=2, figsize=(7, 3))
            for i in range(samples.shape[0]):
                for l in range(self.num_latents):
                    axes[0].plot(softplus(samples[i][:, l]))
        train_y = samples[0]
        self.f_observed_highres = train_y.unsqueeze(0).permute(0, 2, 1)
        self.f_observed = train_y[::self.num_disc+1].unsqueeze(0).permute(0, 2, 1)

        likelihood = MultitaskGaussianLikelihood(num_tasks=self.num_latents)

        mll = VariationalELBO(likelihood, gp_model, num_data=train_y.size(0))
        optimizer = torch.optim.Adam([
            {'params': gp_model.parameters()},
            {'params': likelihood.parameters()},
        ], lr=0.1)

        #train gp_model
        for i in range(self.epochs):
            optimizer.zero_grad()
            output = gp_model(t_predict)
            loss = -mll(output, train_y)
            loss.backward()
            optimizer.step()

        q_f = gp_model(t_predict)
        self.gp_model = gp_model
        samples = q_f.sample(torch.Size([1]))
        if self.plot:
            for i in range(samples.shape[0]):
                for l in range(self.num_latents):
                    axes[1].plot(softplus(samples[i][:, l]))



import torch
import numpy as np
import pandas as pd

from pathlib import Path
from scipy.special import wofz

from alfi.datasets import LFMDataset
from alfi.utilities.data import generate_neural_dataset_2d

PI = torch.tensor(np.pi, requires_grad=False)


class ReactionDiffusion(LFMDataset):
    """
    Reaction diffusion equations with different hyperparameter settings.
    Generated using the `ReactionDiffusionGenerator`
    'l1', 'l2', 'decay', 'diffusion'
    """
    def __init__(self, data_dir='../data/', max_n=2000, nn_format=True, ntest=50, sub=1):
        data = torch.load(Path(data_dir) / 'toydata.pt')
        orig_data = data['orig_data']
        x_observed = data['x_observed']
        self.num_data = min(orig_data.shape[0], max_n)
        self.num_outputs = 1
        self.num_discretised = 40

        tx = x_observed.t()[:, 0:2]
        t_sorted = np.argsort(tx[:, 0], kind='mergesort')
        x_observed = x_observed[:, t_sorted]
        self.orig_data = torch.cat([
            x_observed.unsqueeze(0).repeat(self.num_data, 1, 1),
            orig_data[:self.num_data, :, t_sorted]
        ], dim=1)
        params = data['params'][:self.num_data]

        if nn_format:
            train, test = generate_neural_dataset_2d(self.orig_data, params, self.num_data - ntest, ntest, sub=sub)
            self.data = train
            self.train_data = train
            self.test_data = test
        else:
            self.data = [(x_observed, orig_data[i, :, t_sorted], params[i])
                         for i in range(self.num_data)]

        self.gene_names = np.array(['toy'])


class HomogeneousReactionDiffusion(LFMDataset):
    """
    Homogeneous meaning the decay, sensitivity, diffusion, and lengthscale hyperparameters are all fixed at:
    S=1, l1=0.3, l2=0.3, λ=0.1, D=0.01
    This is the toy dataset from López-Lopera et al. (2019)
    https://arxiv.org/abs/1808.10026
    Data download: https://github.com/anfelopera/PhysicallyGPDrosophila
    nn_format is compatible with neural operator models.
    """
    def __init__(self, data_dir='../data/', one_fixed_sample=True, highres=False, nn_format=None, ntest=50, sub=1):
        if one_fixed_sample:
            data = pd.read_csv(Path(data_dir) / 'demToy1GPmRNA.csv')
            nn_format = False if nn_format is None else nn_format
        else:
            if highres:
                data = pd.read_csv(Path(data_dir) / 'toy_GPmRNA_N50highres.csv')
            else:
                data = pd.read_csv(Path(data_dir) / 'toy_GPmRNA_N1050.csv')
            nn_format = True if nn_format is None else nn_format
        num_per_data = np.unique(data.values[:, 0]).shape[0] * np.unique(data.values[:, 1]).shape[0]
        self.num_data = data.values.shape[0] // num_per_data
        self.orig_data = torch.tensor(data.values).reshape(self.num_data, num_per_data, 4).permute(0, 2, 1)
        self.num_outputs = 1

        tx = self.orig_data[0, 0:2, :]
        t_sorted = np.argsort(tx[0, :], kind='mergesort')
        values = data.values[t_sorted]
        self.orig_data = self.orig_data[..., t_sorted]

        x_observed = torch.tensor(values[:, 0:2]).permute(1, 0)
        num_data = x_observed.shape[1] // num_per_data
        data = torch.tensor(values[:, 3]).unsqueeze(0)
        self.num_discretised = 40
        self.disc = 1
        self.gene_names = np.array(['toy'])

        if nn_format:
            params = torch.tensor([0.3, 0.3, 1.0, 0.1, 0.01]).unsqueeze(0).repeat(self.orig_data.shape[0], 1)
            train, test = generate_neural_dataset_2d(self.orig_data, params, self.num_data - ntest, ntest, sub=sub)
            self.data = train
            self.train_data = train
            self.test_data = test
        else:
            self.data = [
                (x_observed[:, num_per_data*i:num_per_data*(i+1)], data[:, num_per_data*i:num_per_data*(i+1)])
                for i in range(num_data)
            ]


class ReactionDiffusionGenerator:
    """
    Reaction Diffusion LFM generator (transcribed from R to Python by Jacob Moss).
    The original author for the R code for the reaction diffusion equation is López-Lopera et al. (2019)
    https://github.com/anfelopera/PhysicallyGPDrosophila
    """

    def __init__(self, lengthscale=None, sensitivity=1.0, decay=0.1, diffusion=0.01):
        super().__init__()
        l = [0.3, 0.3] if lengthscale is None else lengthscale
        self.lengthscale = torch.tensor(l, dtype=torch.float64)
        self.sensitivity = torch.tensor(sensitivity, dtype=torch.float64)
        self.decay = torch.tensor(decay, dtype=torch.float64)
        self.diffusion = torch.tensor(diffusion, dtype=torch.float64)

    def theta_x(self):
        return self.lengthscale[1]

    def theta_t(self):
        return self.lengthscale[0]

    def Hfun(self, beta_s, t1, t2):
        # t1, t2 = vectors with  (time coordinates)
        # precomputing some terms
        v = 0.5 * self.theta_t() * beta_s
        # diff_t = outer(t2, t1,'-')
        diff_t = t2.view(-1, 1) - t1.view(-1)

        arg1 = v + t1 / self.theta_t()
        arg2 = v - diff_t / self.theta_t()

        # computing the function H for the sim kernel
        return torch.erf(arg1).t() - torch.erf(arg2)

    def hfun(self, t1, t2, beta_s, beta_q):
        # t1, t2 = vectors with  (time coordinates)
        v = 0.5 * self.theta_t() * beta_s

        # diff_t = outer(t2, t1,'-')
        diff_t = t2.view(-1, 1) - t1.view(-1)
        t0 = 0. * t1
        # computing the function h for the sim kernel
        hpart1 = torch.exp(-beta_s * diff_t) * self.Hfun(beta_s, t1, t2)
        thing = (beta_s * t2).view(-1, 1) + (beta_q * t1).view(-1)
        # outer(beta_s*t2,beta_q*t1,'+')

        hpart2 = torch.exp(-thing) * self.Hfun(beta_s, t0, t2)

        h = torch.exp(v ** 2) * (hpart1 - hpart2) / (beta_s + beta_q)
        return h  # .real

    def simXsimKernCompute(self, t1, t2, n, m, l):
        # t1, t2 = vectors with  (time coordinates)
        omega_n2 = (n * np.pi / l) ** 2
        omega_m2 = (m * np.pi / l) ** 2

        beta_q = self.decay + self.diffusion * omega_n2
        beta_s = self.decay + self.diffusion * omega_m2
        # computing the sim kernel
        kern = self.hfun(t1, t2, beta_s, beta_q).t() + \
               self.hfun(t2, t1, beta_q, beta_s)

        kern = 0.5 * np.sqrt(np.pi) * self.theta_t() * kern

        return kern

    def sheatXsheatKernCompute(self, x1, x2, n, m, l):
        # x1, x2 = vectors with  (space coordinates)
        theta_x = self.theta_x()
        omega_n = n * np.pi / l
        gamma_n = 1j * omega_n
        z1_n = 0.5 * theta_x * gamma_n
        z2_n = l / theta_x + z1_n
        wz1_n = wofz(1j * z1_n)
        wz2_n = wofz(1j * z2_n)
        Wox_n = wz1_n - torch.exp(-(l / theta_x) ** 2 - gamma_n * l) * wz2_n
        if n == m:
            omega_m = omega_n
            c = ((theta_x * n * np.pi) ** 2 + 2 * l ** 2) / (2 * np.pi * n * l ** 2)
            c = Wox_n.real - Wox_n.imag * c
            c = 0.5 * theta_x * np.sqrt(np.pi) * l * c + \
                0.5 * theta_x ** 2 * (torch.exp(-(l / theta_x) ** 2) * np.cos(n * np.pi) - 1)
        else:
            omega_m = m * np.pi / l
            gamma_m = 1j * omega_m
            z1_m = 0.5 * theta_x * gamma_m
            z2_m = l / theta_x + z1_m
            wz1_m = wofz(1j * z1_m)
            wz2_m = wofz(1j * z2_m)
            Wox_m = wz1_m - torch.exp(-(l / theta_x) ** 2 - gamma_m * l) * wz2_m
            c = n * Wox_m.imag - m * Wox_n.imag
            c = c * (theta_x * l / (np.sqrt(np.pi) * (m ** 2 - n ** 2)))

        # computing the sheat kernel
        # kern = outer(torch.sin(omega_n*x1),
        #              torch.sin(omega_m*x2),'*') * c
        # print(torch.sin(omega_n*x1).shape, c.shape)
        kern = torch.outer(torch.sin(omega_n * x1),
                           torch.sin(omega_m * x2))
        kern = kern * c
        return kern

    def sheatXrbfKernCompute(self, x1, x2, n, l):
        # x1, x2 = vectors with  (space coordinates)
        # precomputing some terms
        omega_n = n * np.pi / l
        gamma_n = 1j * omega_n
        z1_n = 0.5 * self.theta_x() * gamma_n + x2 / self.theta_x()
        z2_n = z1_n - l / self.theta_x()
        wz1_n = wofz(1j * z1_n)
        wz2_n = wofz(1j * z2_n)
        Wox_n = np.exp(-((x2 - l) / self.theta_x()) ** 2 + gamma_n * l) * wz2_n - \
                np.exp(-(x2 / self.theta_x()) ** 2) * wz1_n
        c = 0.5 * np.sqrt(np.pi) * self.theta_x() * Wox_n.imag

        # computing the sheatXrbf kernel
        # kern = outer(sin(omega_n*x1), c)
        kern = torch.outer(torch.sin(omega_n * x1), c)

        return kern

    def simXrbfKernCompute(self, t1, t2, n, l):
        # t1, t2 = vectors with  (time coordinates)
        # precomputing some terms
        omega_n2 = (n * np.pi / l) ** 2
        beta_q = self.decay + self.diffusion * omega_n2
        z1 = 0.5 * self.theta_t() * beta_q
        # diff_t = outer(t1,t2,'-')
        diff_t = t1.view(-1, 1) - t2.view(-1)
        # computing the simXrbf kernel
        hterm = self.Hfun(beta_q, t2, t1)

        kern = np.exp(z1 ** 2 - beta_q * diff_t) * hterm
        kern = 0.5 * np.sqrt(np.pi) * self.theta_t() * kern
        return kern

    def kyy(self, tx1, tx2):
        nterms = 10
        l = 1
        kern = 0
        cache = dict()
        for i in range(1, nterms + 1):
            for j in range(1, nterms + 1):
                if ((i + j) % 2) == 0:
                    # computing kernel
                    key = f'{j},{i}'
                    if key in cache.keys():
                        kernt = cache[key][0].t()
                        kernx = cache[key][1].t()
                    else:
                        kernt = self.simXsimKernCompute(tx1[:, 0], tx2[:, 0], i, j, l)
                        kernx = self.sheatXsheatKernCompute(tx1[:, 1], tx2[:, 1], i, j, l)
                        key = f'{i},{j}'
                        cache[key] = [kernt, kernx]

                    kern = kern + kernt * kernx

        kern = ((2 * self.sensitivity / l) ** 2) * kern
        return kern

    def kuy(self, x1, x2):
        # x1, x2 = matrices with  (time coordinate, space coordinate)
        # par = (sigma2, theta_t, theta_x, decay (lambda), diffusion (D), sensitivity (S))
        nterms = 10
        l = 1
        kern = 0

        # computing kernel and its gradient
        for i in range(1, nterms + 1):
            # computing kernel
            kernt = self.simXrbfKernCompute(x2[:, 0], x1[:, 0], i, l)
            kernx = self.sheatXrbfKernCompute(x2[:, 1], x1[:, 1], i, l)
            kern = kern + kernt * kernx

        kern = (2 * self.sensitivity / l) * kern.t()

        return kern

    def kuu(self, tx1, tx2):
        t1 = tx1[:, 0]
        t2 = tx2[:, 0]
        x1 = tx1[:, 1]
        x2 = tx2[:, 1]
        theta_t = self.theta_t()
        theta_x = self.theta_x()
        diff_t = t1.view(-1, 1) / theta_t - t2.view(-1) / theta_t
        diff_x = x1.view(-1, 1) / theta_x - x2.view(-1) / theta_x
        dist_t = torch.square(diff_t)
        dist_x = torch.square(diff_x)
        # computing the SE kernel
        return torch.exp(-dist_t - dist_x)

    def joint(self, tx1, tx2):
        tx1 = tx1.type(torch.float64)
        tx2 = tx2.type(torch.float64)
        Kuu = Kyy = Kuy = Kyu = torch.zeros(10)
        # Kuu = self.prior(tx1[:, :2], tx2[:, :2])
        Kuu = self.kuu(tx1, tx2)
        Kyy = self.kyy(tx1, tx2)
        Kuy = self.kuy(tx1, tx2)
        Kyu = Kuy.t()  # self.kyu(tx1, tx2)
        return Kuu, Kyy, Kyu, Kuy



import torch
from torch.nn.functional import softplus


class LpLoss(object):
    def __init__(self, d=2, p=2, size_average=True, reduction=True):
        super(LpLoss, self).__init__()
        #Dimension and Lp-norm type are postive
        assert d > 0 and p > 0

        self.d = d
        self.p = p
        self.reduction = reduction
        self.size_average = size_average

    def abs(self, x, y):
        num_examples = x.size()[0]

        #Assume uniform mesh
        h = 1.0 / (x.size()[1] - 1.0)

        all_norms = (h**(self.d/self.p))*torch.norm(x.view(num_examples,-1) - y.view(num_examples,-1), self.p, 1)

        if self.reduction:
            if self.size_average:
                return torch.mean(all_norms)
            else:
                return torch.sum(all_norms)

        return all_norms

    def rel(self, x, y):
        num_examples = x.size()[0]
        # x = x[..., 0]
        y = y.view(num_examples, -1)
        x_mean = x[..., 0].view(num_examples, -1)
        x_var = softplus(x[..., 1]).view(num_examples, -1) + 1e-6
        # print(x.shape, x_mean.shape, x_var.shape)
        # diff_norms = torch.norm(x_mean.reshape(num_examples,-1) - y.reshape(num_examples,-1), self.p, 1)
        # y_norms = torch.norm(y, self.p, 1)
        # y_norms = y.square().mean(-1).sqrt()
        diff_norms = 0.5 * (torch.log(x_var) + (y - x_mean)**2 / x_var)

        norms = diff_norms.mean(-1)#/ y_norms# #

        if self.reduction:
            if self.size_average:
                return torch.mean(norms)
            else:
                return torch.sum(norms)

        return norms

    def __call__(self, x, y):
        return self.rel(x, y)

"""
        y = y.view(num_examples, -1)
        x_mean = x[..., 0].view(num_examples, -1)
        x_var = softplus(x[..., 1]).view(num_examples, -1) + 1e-6

        diff_norms = torch.norm(x_mean.reshape(num_examples,-1) - y.reshape(num_examples,-1), self.p, 1)
        y_norms = torch.norm(y, self.p, 1)
        # diff_norms = 0.5 * (torch.log(x_var) + (y - x_mean)**2 / x_var)

        norms = diff_norms/ y_norms#.sum(-1) #

        if self.reduction:
            if self.size_average:
                return torch.mean(norms)
            else:
                return torch.sum(norms)

        return norms
"""


import torch
from torch.nn import Parameter
from matplotlib import pyplot as plt
from torch.optim import Adam
from gpytorch.optim import NGD
from gpytorch.constraints import Positive
import seaborn as sns
import numpy as np

from alfi.configuration import VariationalConfiguration
from alfi.models import OrdinaryLFM, generate_multioutput_gp
from alfi.trainers import VariationalTrainer
from gpytorch.mlls.exact_marginal_log_likelihood import ExactMarginalLogLikelihood

from experiments.model_specs.variational import TranscriptionLFM

from alfi.datasets import P53Data
from alfi.models import ExactLFM
from alfi.trainers import ExactTrainer
from collections import namedtuple

""" Experiment for finding timings """


dataset = P53Data(replicate=0, data_dir='./data')
num_genes = 5
num_tfs = 1
config = VariationalConfiguration(
    preprocessing_variance=dataset.variance,
    num_samples=80,
    initial_conditions=False
)

t_end = dataset.t_observed[-1]
num_inducing = 20  # (I x m x 1)

class TempExactTrainer(ExactTrainer):
    def after_epoch(self):
        super().after_epoch()
        with torch.no_grad():
            sens = self.lfm.covar_module.sensitivity
            sens[3] = np.float64(1.)
            deca = self.lfm.covar_module.decay
            deca[3] = np.float64(0.8)
            self.lfm.covar_module.sensitivity = sens
            self.lfm.covar_module.decay = deca

class TempVariationalTrainer(VariationalTrainer):
    def after_epoch(self):
        with torch.no_grad():
            sens = torch.tensor(1.)
            dec = torch.tensor(0.8)
            self.lfm.raw_sensitivity[3] = self.lfm.positivity.inverse_transform(sens)
            self.lfm.raw_decay[3] = self.lfm.positivity.inverse_transform(dec)
        super().after_epoch()

if __name__ == '__main__':
    with open('experiments/times.txt', 'w') as f:
        num_genes = 5
        num_tfs = 1
        config = VariationalConfiguration(
            num_samples=80,
            initial_conditions=False
        )
        exact_times_list = list()
        varia_times_list = list()

        for i in range(10):
            print('Running ', i)
            #### Train Exact
            exact_lfm = ExactLFM(dataset, dataset.variance.reshape(-1))
            optimizer = torch.optim.Adam(exact_lfm.parameters(), lr=0.01)

            loss_fn = ExactMarginalLogLikelihood(exact_lfm.likelihood, exact_lfm)


            exact_trainer = TempExactTrainer(
                exact_lfm, [optimizer], dataset, loss_fn=loss_fn)

            exact_lfm.train()
            exact_lfm.likelihood.train()

            exact_times = exact_trainer.train(epochs=600, report_interval=50)
            exact_times = np.array(exact_times)
            exact_times_list.append(exact_times)

            #### Train variational

            inducing_points = torch.linspace(0, t_end, num_inducing).repeat(num_tfs, 1).view(num_tfs, num_inducing, 1)
            gp_model = generate_multioutput_gp(num_tfs, inducing_points, gp_kwargs=dict(natural=True))

            lfm = TranscriptionLFM(num_genes, gp_model, config)

            track_parameters = [
                'raw_basal',
                'raw_decay',
                'raw_sensitivity',
                'gp_model.covar_module.raw_lengthscale',
            ]
            num_training = dataset.m_observed.shape[-1]

            variational_optimizer = NGD(lfm.variational_parameters(), num_data=num_training, lr=0.1)
            parameter_optimizer = Adam(lfm.nonvariational_parameters(), lr=0.03)
            optimizers = [variational_optimizer, parameter_optimizer]
            trainer = TempVariationalTrainer(lfm, optimizers, dataset, track_parameters=track_parameters)

            lfm.train()
            varia_times = trainer.train(600, step_size=5e-1, report_interval=50)
            varia_times = np.array(varia_times)
            varia_times_list.append(varia_times)
            last_loss = trainer.losses[-1].sum()

        varia_times = np.stack(varia_times_list, axis=0)
        exact_times = np.stack(exact_times_list, axis=0)

        np.save('./experiments/exact_times.npy', exact_times)
        np.save('./experiments/varia_times.npy', varia_times)

        print('finished')



import argparse
import yaml
import seaborn as sns
import time
import numpy as np
import torch
import matplotlib as mpl
mpl.use('Agg')

from matplotlib import pyplot as plt
from pathlib import Path

from alfi.datasets import (
    P53Data, HafnerData, ToyTranscriptomicGenerator,
    HomogeneousReactionDiffusion, DrosophilaSpatialTranscriptomics,
    DeterministicLotkaVolterra, ReactionDiffusion
)
from alfi.utilities.torch import get_mean_trace, is_cuda
from alfi.models import TrainMode
try:
    from .model_specs.partial import build_partial, plot_partial, pretrain_partial
except ImportError:
    build_partial, plot_partial, pretrain_partial = None, None, None


from .model_specs.variational import build_variational, plot_variational
from .model_specs.exact import build_exact, plot_exact
from .model_specs.lotka import build_lotka, plot_lotka
from .model_specs.lfo import build_dataset, build_lfo

"""
MAIN EXPERIMENT SCRIPT
This script is used to run the main experiments to test Alfi. Further scripts for calculating the errors displayed
in the tables in the Alfi paper can be found in the `evaluation_scripts_for_paper` directory.

It may be clearer to start with the Jupyter notebooks found in docs/notebooks, as they go through some experiments
step-by-step.
"""

plt.rcParams['font.family'] = 'serif'
plt.rcParams['font.serif'] = 'CMU Serif'
sns.set(font="CMU Serif")
class TerminalColours:
    GREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    END = '\033[0m'

# ------Config------ #

with open("experiments/experiments.yaml", 'r') as stream:
    try:
        config = yaml.safe_load(stream)
    except yaml.YAMLError as exc:
        print(exc)
        exit()

dataset_choices = list(config.keys())

parser = argparse.ArgumentParser()
parser.add_argument('--data', type=str, choices=dataset_choices, default=dataset_choices[0])
parser.add_argument('--reload', type=bool, default=False)
parser.add_argument('--timer', type=bool, default=False)
parser.add_argument('--timer_samples', type=int, default=5)
parser.add_argument('--experiment_index', type=int, default=-1)

datasets = {
    'p53': lambda: P53Data(replicate=0, data_dir='data'),
    'hafner': lambda: HafnerData(replicate=0, data_dir='data', extra_targets=False),
    'toy-spatial': lambda: HomogeneousReactionDiffusion(data_dir='data'),
    'dros-kr': lambda: DrosophilaSpatialTranscriptomics(gene='kr', data_dir='data', scale=True),
    'toy': lambda: ToyTranscriptomicGenerator().generate_single(),
    'lotka': lambda: DeterministicLotkaVolterra(alpha=2./3, beta=4./3,
                                                gamma=1., delta=1.,
                                                steps=13, end_time=12, fixed_initial=0.8),
    'reaction-diffusion': lambda: build_dataset(
        ReactionDiffusion('data', nn_format=True, max_n=4000, ntest=50), ntest=50
    ),
}


# ------Set up model initialisers------ #
builders = {
    'variational': build_variational,
    'exact': build_exact,
    'partial': build_partial,
    'lotka': build_lotka,
    'lfo-2d': lambda *args, **kwargs: build_lfo(*args, **kwargs, block_dim=2),
}

plotters = {
    'exact': plot_exact,
    'partial': plot_partial,
    'variational': plot_variational,
    'lotka': plot_lotka,
}


train_pre_step = {
    'exact': lambda _, model, __, ___: model.likelihood.train(),
    'partial': pretrain_partial
}


def time_models(builder, dataset, filepath, modelparams, num_samples):
    times_with = list()
    loglosses_with = list()
    times_without = list()
    loglosses_without = list()

    for i in range(num_samples):
        # Without pretraining
        print(TerminalColours.GREEN, 'Without pretraining...', TerminalColours.END)
        model, trainer, plotter = builder(dataset, modelparams)
        trainer.plot_outputs = False

        t0 = time.time()
        model.set_mode(TrainMode.NORMAL)
        train_times = trainer.train(**experiment['train_params'])
        train_times = np.array(train_times)
        train_time = (train_times[:, 0] - t0) / 60
        logloss = train_times[:, 1]
        times_without.append(train_time)
        loglosses_without.append(logloss)
        model.save(str(filepath / f'model_without_{i}'))
        torch.save(get_mean_trace(trainer.parameter_trace), filepath / f'parameter_trace_without_{i}.pt')
        # With pretraining
        print(TerminalColours.GREEN, 'With pretraining...', TerminalColours.END)
        model, trainer, plotter = builder(dataset, modelparams)
        trainer.plot_outputs = False

        model.set_mode(TrainMode.GRADIENT_MATCH)
        pretrain_times, t_start = train_pre_step[method](dataset, model, trainer, modelparams)

        t1 = time.time()
        model.set_mode(TrainMode.NORMAL)
        train_times = trainer.train(**experiment['train_params'])
        pretrain_times = np.array(pretrain_times)
        train_times = np.array(train_times)
        pretrain_time = (pretrain_times[:, 0] - t_start) / 60
        train_time = (train_times[:, 0] - t1 + pretrain_time[-1]) / 60
        times_with.append(np.concatenate([pretrain_time, train_time]))
        loglosses_with.append(np.concatenate([pretrain_times[:, 1], train_times[:, 1]]))

        model.save(str(filepath / f'model_with_{i}'))
        torch.save(get_mean_trace(trainer.parameter_trace), filepath / f'parameter_trace_with_{i}.pt')

    loglosses_with = np.array(loglosses_with)
    loglosses_without = np.array(loglosses_without)
    times_with = np.array(times_with)
    times_without = np.array(times_without)
    np.save(str(filepath / 'time_with.npy'), times_with)
    np.save(str(filepath / 'loss_with.npy'), loglosses_with)
    np.save(str(filepath / 'time_without.npy'), times_without)
    np.save(str(filepath / 'loss_without.npy'), loglosses_without)


def run_model(method, dataset, model, trainer, plotter, filepath, save_filepath, modelparams):
    # Train model with optional initial step
    if method in train_pre_step:
        train_pre_step[method](dataset, model, trainer, modelparams)
    print(TerminalColours.GREEN, 'Training...', TerminalColours.END)
    times = trainer.train(**experiment['train_params'])
    np.save(save_filepath + '_times.npy', times)

    # Plot results of model
    if method in plotters:
        print(TerminalColours.GREEN, 'Running plotter...', TerminalColours.END)
        plotters[method](dataset, model, trainer, plotter, filepath, modelparams)
    else:
        print(TerminalColours.WARNING, 'Ignoring plotter for', method, 'since no plotter implemented.',
              TerminalColours.END)
    model.save(save_filepath)


if __name__ == "__main__":
    args = parser.parse_args()
    key = args.data
    experiment_index = args.experiment_index

    print(TerminalColours.GREEN, 'Running experiments for dataset:', key, TerminalColours.END)
    data_config = config[key]
    dataset = datasets[key]()
    experiments = data_config['experiments']
    seen_methods = dict()
    for i, experiment in enumerate(experiments):
        if experiment_index > -1 and experiment_index != i:
            continue
        method = experiment['method']
        print(TerminalColours.GREEN, 'Constructing method:', method, TerminalColours.END)
        if method in builders:
            if method not in seen_methods:
                seen_methods[method] = 0

            # Create experiments path
            filepath = Path('experiments', key, method)
            filepath.mkdir(parents=True, exist_ok=True)
            save_filepath = str(filepath / (str(seen_methods[method]) + 'savedmodel'))

            # Construct model
            modelparams = experiment['model-params'] if 'model-params' in experiment else None
            reload = save_filepath if args.reload else None
            model, trainer, plotter = builders[method](
                dataset, modelparams, reload=reload, checkpoint_dir=filepath)
            model = model.cuda() if is_cuda() else model
            if args.timer:
                time_models(builders[method], dataset, filepath, modelparams, args.timer_samples)
            else:
                run_model(method, dataset, model, trainer, plotter, filepath, save_filepath, modelparams)
            seen_methods[method] += 1
            print(TerminalColours.GREEN, f'{method} completed successfully.', TerminalColours.END)
        else:
            print(TerminalColours.WARNING, 'Ignoring method', method, 'since no builder implemented.', TerminalColours.END)



import argparse
import yaml
import seaborn as sns
import time
import numpy as np
import torch

from matplotlib import pyplot as plt
from pathlib import Path

from alfi.datasets import P53Data
from alfi.plot import tight_kwargs

from experiments.model_specs.variational import build_variational, plot_variational

"""
This experiment compares the parameters on which Alfi converges with those the original Barenco paper found.
"""
# ------Config------ #

with open("experiments/experiments.yaml", 'r') as stream:
    try:
        config = yaml.safe_load(stream)
    except yaml.YAMLError as exc:
        print(exc)
        exit()

dataset_choices = list(config.keys())

parser = argparse.ArgumentParser()
parser.add_argument('--samples', type=int, default=5)


if __name__ == "__main__":
    args = parser.parse_args()
    data_config = config['p53']
    experiments = data_config['experiments']
    for experiment in experiments:
        if experiment['method'] == 'variational':
            break

    print(experiment)
    method = experiment['method']

    # Create experiments path
    filepath = Path('experiments', 'p53', method)
    filepath.mkdir(parents=True, exist_ok=True)

    kinetics_samples = list()
    for i in range(args.samples):
        dataset = P53Data(replicate=0, data_dir='data')

        # Construct model
        modelparams = experiment['model-params'] if 'model-params' in experiment else None
        model, trainer, plotter = build_variational(
            dataset, modelparams, checkpoint_dir=None)

        trainer.train(**experiment['train_params'], step_size=5e-1)

        kinetics = np.array(
            [x.detach().squeeze().numpy() for x in [model.basal_rate, model.sensitivity, model.decay_rate]])

        kinetics_samples.append(kinetics)

    samples = np.stack(kinetics_samples)
    print(samples.shape)
    kinetics = samples.mean(0)
    err = samples.std(0)
    print(kinetics.shape, err.shape)
    labels = ['Basal rates', 'Sensitivities', 'Decay rates']

    plotter.plot_double_bar(kinetics, labels, params_var=err, ground_truths=P53Data.params_ground_truth(),
                            figsize=(6.5, 2.3),
                            yticks=[
                                np.linspace(0, 0.12, 5),
                                np.linspace(0, 1.2, 4),
                                np.arange(0, 1.1, 0.2),
                            ])
    plt.tight_layout()

    plt.savefig(filepath / 'linear-kinetics2.pdf', **tight_kwargs)



import torch
from torch.nn import Parameter
from matplotlib import pyplot as plt
from torch.optim import Adam
from gpytorch.optim import NGD
import numpy as np

from alfi.datasets import P53Data, ToyTranscriptomicGenerator
from alfi.configuration import VariationalConfiguration
from alfi.models import OrdinaryLFM, generate_multioutput_gp
from alfi.trainers import VariationalTrainer
from .variational import TranscriptionLFM
from alfi.utilities.torch import q2
from torch.nn.functional import l1_loss
from gpytorch.mlls.exact_marginal_log_likelihood import ExactMarginalLogLikelihood

from alfi.models import ExactLFM
from alfi.trainers import ExactTrainer
from collections import namedtuple

"""
Experiment for finding mean absolute errors in the toy transcriptomic dataset
"""
def get_datasets(data_dir='.'):
    from alfi.datasets import LFMDataset
    class DSet(LFMDataset):
        def __init__(self, disc, t_obs, m_obs, f_obs, lfm):
            self.num_outputs = 5
            self.m_observed = m_obs[:, :, ::disc]
            self.m_observed_highres = m_obs

            self.f_observed = f_obs[:, :, ::disc]
            self.f_observed_highres = f_obs

            self.t_observed = t_obs[::disc]
            self.t_observed_highres = t_obs
            self.lfm = lfm
            self.data = [(self.t_observed, self.m_observed[0, i]) for i in range(self.num_outputs)]
            self.gene_names = np.arange(self.num_outputs)

    datasets = list()
    for i in range(10):
        dataset_dict = torch.load(f'{data_dir}/experiments/dataset{i}.pt')
        lfm = namedtuple('lfm', ['basal_rate', 'sensitivity', 'decay_rate'])
        lfm = lfm(dataset_dict['basal'], dataset_dict['sensitivity'], dataset_dict['decay'])
        dataset = DSet(
            dataset_dict['high_disc'],
            dataset_dict['t_obs'],
            dataset_dict['m_obs'],
            dataset_dict['f_obs'], lfm)
        datasets.append(dataset)
    return datasets

def save_dataset(dataset, path):
    torch.save({
        'm_obs': dataset.m_observed_highres, 'high_disc': 10,
        't_obs': dataset.t_observed_highres,
        'f_obs': dataset.f_observed_highres,
        'lengthscale': dataset.lfm.lengthscale,
        'basal': dataset.lfm.basal_rate,
        'sensitivity': dataset.lfm.sensitivity,
        'decay': dataset.lfm.decay_rate
    }, path)


def get_params(lfm):
    if isinstance(lfm, ExactLFM):
        basal = lfm.mean_module.basal
        sensitivity = lfm.covar_module.sensitivity
        decay = lfm.covar_module.decay
    else:
        basal = lfm.basal_rate.squeeze()
        sensitivity = lfm.sensitivity.squeeze()
        decay = lfm.decay_rate.squeeze()
    return basal, sensitivity, decay

def param_l1(lfm, ground_truth):
    basal, sensitivity, decay = get_params(lfm)
    return (
        l1_loss(basal, torch.from_numpy(ground_truth[0])) +
        l1_loss(sensitivity, torch.from_numpy(ground_truth[1])) +
        l1_loss(decay, torch.from_numpy(ground_truth[2]))
    ).mean().item()

def train(lfm, trainer, m_targ, f_targ, epochs=200, **train_kwargs):
    f_maes = list()
    m_maes = list()
    f_q2s = list()
    m_q2s = list()
    i = 0
    for i in range(epochs // 10):
        try:
            trainer.train(epochs=10, report_interval=50, **train_kwargs)
        except:
            break
        m_pred = lfm.predict_m(t_predict, jitter=1e-3)
        f_pred = lfm.predict_f(t_predict, jitter=1e-3)
        m_mae = l1_loss(m_pred.mean, m_targ).mean().item()
        f_mae = l1_loss(f_pred.mean, f_targ).mean().item()
        f_q2 = q2(f_pred.mean, f_targ).mean().item()
        m_q2 = q2(m_pred.mean, m_targ).mean().item()
        f_maes.append(f_mae)
        m_maes.append(m_mae)
        f_q2s.append(f_q2)
        m_q2s.append(m_q2)
    if i < 10:
        print('except')
        raise Exception('too few epochs')
    f_maes = torch.tensor(f_maes)
    m_maes = torch.tensor(m_maes)
    min_index = (f_maes + m_maes).argmin(dim=0)
    f_mae = f_maes[min_index]
    m_mae = m_maes[min_index]

    f_q2s = torch.tensor(f_q2s)
    m_q2s = torch.tensor(m_q2s)
    min_index = (f_q2s + m_q2s).argmax(dim=0)
    f_q2 = f_q2s[min_index]
    m_q2 = m_q2s[min_index]

    return m_mae, f_mae, m_q2, f_q2


if __name__ == '__main__':
    generate = False
    if not generate:
        datasets = get_datasets()
    with open('experiments/mae.txt', 'w') as f:
        f.write(
            f'iter\tm_mae_exact\tm_mae_varia\tf_mae_exact\tf_mae_varia\tm_q2_exact'
            f'\tm_q2_varia\tf_q2_exact\tf_q2_varia\tparam_mae_exact\tparam_mae_varia\n')

        num_genes = 5
        num_tfs = 1
        config = VariationalConfiguration(
            num_samples=80,
            initial_conditions=False
        )
        maes = list()
        q2s = list()
        num_success = 0
        for i in range(20):
            if num_success >= 10:
                break
            print('Running ', i, num_success)
            if generate:
                exit()
                dataset = ToyTranscriptomicGenerator(
                    num_outputs=5,
                    num_latents=1,
                    num_times=12,
                    softplus=False,
                    dtype=torch.float32)
                dataset.generate_single(lengthscale=1.5)
            else:
                if i >= len(datasets):
                    break
                dataset = datasets[i]

            dataset.variance = 1e-3 * torch.ones(dataset.m_observed.shape[-1], dtype=torch.float32)
            start_time = dataset.t_observed[0]
            end_time = dataset.t_observed[-1]
            t_predict = torch.linspace(0, end_time, dataset.m_observed_highres.shape[-1], dtype=torch.float32)
            ground_truth = [
                dataset.lfm.basal_rate.detach().view(-1).numpy(),
                dataset.lfm.sensitivity.detach().view(-1).numpy(),
                dataset.lfm.decay_rate.detach().view(-1).numpy()
            ]
            m_targ = dataset.m_observed_highres.squeeze().t()
            f_targ = dataset.f_observed_highres.squeeze(0).t()

            #### Train Exact
            exact_lfm = ExactLFM(dataset, dataset.variance.reshape(-1))
            optimizer = torch.optim.Adam(exact_lfm.parameters(), lr=0.01)

            loss_fn = ExactMarginalLogLikelihood(exact_lfm.likelihood, exact_lfm)

            class ConstrainedTrainer(ExactTrainer):
                def after_epoch(self):
                    with torch.no_grad():
                        sens_fixed = dataset.lfm.sensitivity[0].squeeze().numpy()
                        dec_fixed = dataset.lfm.decay_rate[0].squeeze()
                        sens = self.lfm.covar_module.sensitivity
                        sens[0] = np.float64(sens_fixed)
                        deca = self.lfm.covar_module.decay
                        deca[0] = np.float64(dec_fixed)
                        self.lfm.covar_module.sensitivity = sens
                        self.lfm.covar_module.decay = deca
                    super().after_epoch()


            exact_trainer = ConstrainedTrainer(
                exact_lfm, [optimizer], dataset, loss_fn=loss_fn,
                track_parameters=[
                    'mean_module.raw_basal',
                    'covar_module.raw_decay',
                    'covar_module.raw_sensitivity',
                    'covar_module.raw_lengthscale'
                ])

            exact_lfm.train()
            exact_lfm.likelihood.train()

            try:
                m_mae_exact, f_mae_exact,  m_q2_exact, f_q2_exact = train(exact_lfm, exact_trainer, m_targ, f_targ, epochs=700)
            except Exception as e:
                print('Exception:', e)
                # raise e
                continue

            #### Train variational
            num_inducing = 15  # (I x m x 1)
            inducing_points = torch.linspace(0, end_time, num_inducing).repeat(num_tfs, 1).view(num_tfs, num_inducing, 1)
            gp_model = generate_multioutput_gp(num_tfs, inducing_points, gp_kwargs=dict(natural=True))

            lfm = TranscriptionLFM(num_genes, gp_model, config)

            num_training = dataset.m_observed.shape[-1]

            variational_optimizer = NGD(lfm.variational_parameters(), num_data=num_training, lr=0.06)
            parameter_optimizer = Adam(lfm.nonvariational_parameters(), lr=0.03)
            optimizers = [variational_optimizer, parameter_optimizer]

            class ConstrainedTrainer(VariationalTrainer):
                def after_epoch(self):
                    with torch.no_grad():
                        sens = dataset.lfm.sensitivity[0].squeeze()
                        dec = dataset.lfm.decay_rate[0].squeeze()
                        self.lfm.raw_sensitivity[0] = self.lfm.positivity.inverse_transform(sens)
                        self.lfm.raw_decay[0] = self.lfm.positivity.inverse_transform(dec)
                    super().after_epoch()


            trainer = ConstrainedTrainer(lfm, optimizers, dataset, track_parameters=[
                'raw_basal',
                'raw_decay',
                'raw_sensitivity',
                'gp_model.covar_module.raw_lengthscale',
            ])

            lfm.train()
            try:
                m_mae_varia, f_mae_varia, m_q2_varia, f_q2_varia = train(lfm, trainer, m_targ, f_targ, epochs=700, step_size=5e-1)
            except Exception as e:
                print('exception', e)
                continue
            last_loss = trainer.losses[-1].sum()

            param_mae_exact = param_l1(exact_lfm, ground_truth)
            param_mae_varia = param_l1(lfm, ground_truth)
            print(m_mae_exact, m_mae_varia, f_mae_exact, f_mae_varia)
            maes.append(np.stack([m_mae_exact, m_mae_varia, f_mae_exact, f_mae_varia, param_mae_exact, param_mae_varia]))
            q2s.append(np.stack([m_q2_exact, m_q2_varia, f_q2_exact, f_q2_varia]))
            f.write(f'{i}\t{m_mae_exact}\t{m_mae_varia}\t{f_mae_exact}\t{f_mae_varia}\t{m_q2_exact}\t{m_q2_varia}'
                    f'\t{f_q2_exact}\t{f_q2_varia}\t{param_mae_exact}\t{param_mae_varia}\n')
            if generate:
                save_dataset(dataset, f'./experiments/dataset{num_success}.pt')
            num_success += 1
        maes = np.array(maes)
        q2s = np.array(q2s)
        np.save('./experiments/maes.npy', maes)
        np.save('./experiments/q2s.npy', q2s)
        print(maes)
        print(q2s)
        print(maes.shape)

        print('mae avg', maes.mean(axis=0), maes.std(axis=0))
        print('q2 avg', q2s.mean(axis=0), q2s.std(axis=0))



import numpy as np
import seaborn as sns

from abc import ABC
from matplotlib import pyplot as plt

from .colours import Colours


plt.style.use('seaborn')
plt.rcParams['font.family'] = 'serif'
plt.rcParams['font.serif'] = 'CMU Serif'
sns.set(style='white', font="CMU Serif")


class Plotter(ABC):

    def __init__(self, model, output_names, style='seaborn'):
        self.model = model
        self.output_names = output_names

        plt.style.use(style)
        sns.set(font="CMU Serif", style='white')
        plt.rcParams['font.family'] = 'serif'
        plt.rcParams['font.serif'] = 'CMU Serif'

    def plot_double_bar(self, params_mean,
                        labels=None,
                        titles=None,
                        params_var=None,
                        ground_truths=None,
                        figsize=(8.5, 3),
                        yticks=None,
                        max_plots=10):
        if labels is None:
            labels = self.output_names
        if titles is None:
            titles = self.output_names
        if ground_truths is None:
            ground_truths = np.nan * np.ones_like(params_mean)
        if params_var is None:
            params_var = [None] * len(params_mean)
        num_plots = min(max_plots, params_mean.shape[0])
        fig, axes = plt.subplots(ncols=num_plots, figsize=figsize)
        axes = [axes] if num_plots < 2 else axes
        plotnum = 0
        num_bars = params_mean.shape[-1]
        num_bars = min(max_plots, num_bars)
        for predicted, target, var, title in zip(params_mean, ground_truths, params_var, titles):
            if any(np.isnan(target)):
                axes[plotnum].bar(np.arange(num_bars), predicted[:num_bars],
                                  width=0.4,
                                  tick_label=labels[:num_bars],
                                  color=Colours.bar1_color)
            else:
                axes[plotnum].bar(np.arange(num_bars) - 0.2, predicted[:num_bars],
                                  width=0.4,
                                  tick_label=labels[:num_bars],
                                  color=Colours.bar1_color,
                                  yerr=var[:num_bars] if params_var[0] is not None else None,
                                  capsize=2)
                axes[plotnum].bar(np.arange(num_bars) + 0.2, target[:num_bars],
                                  width=0.4,
                                  color=Colours.bar2_color,
                                  align='center')

            axes[plotnum].set_title(title)
            axes[plotnum].tick_params(axis='x', labelrotation=35)
            if yticks is not None:
                axes[plotnum].set_yticks(yticks[plotnum])
            if num_bars == 1:
                axes[plotnum].set_xlim(-1, 1)
            plotnum += 1

    def plot_losses(self, trainer, last_x=50):
        plt.figure(figsize=(5, 2))
        plt.plot(np.sum(trainer.losses, axis=1)[-last_x:])
        plt.title('Total loss')
        plt.figure(figsize=(5, 2))
        plt.subplot(221)
        plt.plot(trainer.losses[-last_x:, 0])
        plt.title('Loss')
        plt.subplot(222)
        plt.plot(trainer.losses[-last_x:, 1])
        plt.title('KL-divergence')



import gpytorch
import torch

from gpytorch.constraints import Positive


class SIMMean(gpytorch.means.Mean):

    def __init__(self, covar_module, num_genes, initial_basal):
        super().__init__()
        self.covar_module = covar_module
        self.pos_contraint = Positive()
        self.covar_module = covar_module
        self.num_genes = num_genes

        self.register_parameter(
            name='raw_basal', parameter=torch.nn.Parameter(
                self.pos_contraint.inverse_transform(0.05 * torch.ones(self.num_genes)))
        )
        self.register_constraint("raw_basal", self.pos_contraint)

        # self.basal = initial_basal

    @property
    def basal(self):
        return self.pos_contraint.transform(self.raw_basal)

    @basal.setter
    def basal(self, value):
        self.initialize(raw_basal=self.pos_contraint.inverse_transform(value))

    def forward(self, x):
        block_size = int(x.shape[0] / self.num_genes)
        m = (self.basal / self.covar_module.decay).view(-1, 1)
        m = m.repeat(1, block_size).view(-1)
        return m



import torch

from typing import List

from torch.distributions import Normal
from torch.optim.lr_scheduler import StepLR
from torch.optim import Optimizer
from torch.nn.functional import mse_loss, softplus

from alfi.nn import LpLoss
from alfi.utilities.torch import is_cuda
from alfi.models import NeuralOperator
from alfi.trainers import Trainer


class NeuralOperatorTrainer(Trainer):
    """
    Parameters:
        batch_size: in the case of the transcriptional regulation model, we train the entire gene set as a batch
    """
    def __init__(self,
                 lfm: NeuralOperator,
                 optimizers: List[Optimizer],
                 train_loader,
                 test_loader,
                 warm_variational=-1,
                 **kwargs):
        super().__init__(lfm, optimizers, train_loader.dataset, **kwargs)
        self.warm_variational = warm_variational
        self.optimizer = optimizers[0]
        self.train_loader = train_loader
        self.test_loader = test_loader
        gamma = 0.5
        step_size = 100

        self.scheduler = StepLR(self.optimizer, step_size=step_size, gamma=gamma)

        self.loss_fn = LpLoss(size_average=True)

    def to_normal(self, p_y_pred):
        mu = p_y_pred[..., 0]
        sigma = softplus(p_y_pred[..., 1]) + 1e-6
        return Normal(mu, sigma)

    def single_epoch(self, step_size=1e-1, epoch=0, **kwargs):
        self.model.train()
        train_mse = 0
        train_l2 = 0
        train_loss = 0
        train_params_mse = 0
        batch_size = self.train_loader.batch_size
        for batch in self.train_loader:
            if self.model.params:
                x, y, params = batch
                if is_cuda():
                    params = params.cuda()
            else:
                x, y = batch
            if is_cuda():
                x, y = x.cuda(), y.cuda()

            self.optimizer.zero_grad()
            additional_loss = 0

            p_y_pred = self.model(x)
            if self.model.params:
                p_y_pred, params_out = p_y_pred
                params_mse = mse_loss(params_out, params.view(batch_size, -1), reduction='mean')
                train_params_mse += params_mse.item()
                additional_loss = 10 * params_mse

            p_y_pred = self.to_normal(p_y_pred)

            mse = mse_loss(p_y_pred.mean, y.squeeze(-1), reduction='mean')
            l2 = self._loss(p_y_pred, y.squeeze(-1))
            total_loss = l2 + additional_loss

            total_loss.backward()  # use the l2 relative loss
            self.optimizer.step()

            train_mse += mse.item()
            train_l2 += l2.item()
            train_loss += total_loss.item()

        self.scheduler.step()
        self.model.eval()

        test_l2 = 0.0
        test_mse = 0.0
        test_params_mse = 0.0
        test_loss = 0.0
        with torch.no_grad():
            for batch in self.test_loader:
                if self.model.params:
                    x, y, params = batch
                else:
                    x, y = batch
                # x, y = x.cuda(), y.cuda()
                if is_cuda():
                    x, y, params = x.cuda(), y.cuda(), params.cuda()
                additional_loss = 0

                p_y_pred = self.model(x)
                if self.model.params:
                    p_y_pred, params_out = p_y_pred
                    params_mse = mse_loss(params_out, params.view(self.test_loader.batch_size, -1), reduction='mean')
                    test_params_mse += params_mse.item()
                    additional_loss = 10 * params_mse

                p_y_pred = self.to_normal(p_y_pred)

                mse = mse_loss(p_y_pred.mean, y.squeeze(-1), reduction='mean')
                l2 = self._loss(p_y_pred, y.squeeze(-1))
                total_loss = l2 + additional_loss

                test_mse += mse.item()
                test_l2 += l2.item()
                test_loss += total_loss.item()

        train_mse /= len(self.train_loader)
        train_l2 /= len(self.train_loader)
        train_params_mse /= len(self.train_loader)
        train_loss /= len(self.train_loader)
        test_l2 /= len(self.test_loader)
        test_mse /= len(self.test_loader)
        test_params_mse /= len(self.test_loader)
        test_loss /= len(self.test_loader)

        return train_loss, (test_loss, train_l2, test_l2, train_params_mse, test_params_mse)

    def _loss(self, p_y_pred, y_target):
        """
        Computes Neural Operator loss.

        Parameters
        ----------
        p_y_pred : one of torch.distributions.Distribution
            Distribution over y output by Neural Process.

        y_target : torch.Tensor
            Shape (batch_size, num_target, y_dim)

        """
        nll = -p_y_pred.log_prob(y_target).mean()
        return nll



import tensorflow_probability as tfp
import tensorflow as tf
from .mixins import ParamGroupMixin


class HMCSampler(tfp.mcmc.NoUTurnSampler, ParamGroupMixin):

    def __init__(self, likelihood_fn, params_list, step_size):
        self.likelihood_fn = likelihood_fn
        self.param_group = params_list
        self.transforms = [param.transform for param in params_list]
        self.priors = [param.prior for param in params_list]
        super().__init__(self.log_prob, step_size)

    def log_prob(self, *args):
        new_prob = 0

        param_kwargs = {}
        for i, param in enumerate(self.param_group):
            val = self.transforms[i](args[i])
            # tf.print(args[i], val)
            # Prepare likelihood params
            param_kwargs[param.name] = val

            # Add prior:
            new_prob += tf.reduce_sum(self.priors[i].log_prob(val))

        # Add likelihood:
        new_prob += tf.reduce_sum(self.likelihood_fn(**param_kwargs))
        # tf.print(self.param_group[0].name, self.transforms[0](args[0]), new_prob)

        return new_prob




import torch
import numpy as np

from torch.nn import Parameter
from gpytorch.optim import NGD
from torch.optim import Adam
from matplotlib import pyplot as plt
from gpytorch.constraints import Positive, Interval
from gpytorch.distributions import MultitaskMultivariateNormal, MultivariateNormal

from alfi.configuration import VariationalConfiguration
from alfi.models import OrdinaryLFM, MultiOutputGP, generate_multioutput_gp
from alfi.plot import Plotter1d, Colours, tight_kwargs
from alfi.trainers import VariationalTrainer, PreEstimator
from alfi.utilities.data import hafner_ground_truth
from alfi.datasets import HafnerData
from alfi.utilities.torch import softplus
from experiments.variational import TranscriptionLFM


plt.rcParams.update({'font.size': 82})
plt.rc('font', size=15)          # controls default text sizes
plt.rc('axes', titlesize=15)     # fontsize of the axes title
plt.rc('axes', labelsize=15)    # fontsize of the x and y labels
plt.rc('xtick', labelsize=14)    # fontsize of the tick labels
plt.rc('ytick', labelsize=14)    # fontsize of the tick labels
plt.rc('legend', fontsize=20)    # legend fontsize
plt.rc('figure', titlesize=20)  # fontsize of the figure title


class ConstrainedTrainer(VariationalTrainer):
    def after_epoch(self):
        with torch.no_grad():
            sens = torch.tensor(4.2)
            dec = torch.tensor(0.21)
            self.lfm.raw_sensitivity[6] = self.lfm.positivity.inverse_transform(sens)
            self.lfm.raw_decay[6] = self.lfm.positivity.inverse_transform(dec)
        super().after_epoch()


class SoftplusTranscriptionLFM(TranscriptionLFM):
    def __init__(self, num_outputs, gp_model, config: VariationalConfiguration, **kwargs):
        super().__init__(num_outputs, gp_model, config, **kwargs)
        self.positivity = Positive()
        # self.decay_constraint = Interval(0, 1)
        self.raw_decay = Parameter(0.1 + torch.rand(torch.Size([self.num_outputs, 1]), dtype=torch.float64))
        self.raw_basal = Parameter(torch.rand(torch.Size([self.num_outputs, 1]), dtype=torch.float64))
        self.raw_sensitivity = Parameter(8 + torch.rand(torch.Size([self.num_outputs, 1]), dtype=torch.float64))

    def mix(self, f):
        # I = 1 so just repeat for num_outputs
        return softplus(f).repeat(1, self.num_outputs, 1)

    def predict_f(self, t_predict, **kwargs):
        # Sample from the latent distribution
        q_f = super().predict_f(t_predict)
        f = q_f.sample(torch.Size([500])).permute(0, 2, 1)  # (S, I, T)
        print(f.shape)
        # This is a hack to wrap the latent function with the nonlinearity. Note we use the same variance.
        f = torch.mean(self.mix(f), dim=0)[0].unsqueeze(0)
        print(f.shape, q_f.mean.shape, q_f.scale_tril.shape)
        batch_mvn = MultivariateNormal(f, q_f.covariance_matrix.unsqueeze(0))
        print(batch_mvn)
        return MultitaskMultivariateNormal.from_batch_mvn(batch_mvn, task_dim=0)


dataset = HafnerData(replicate=0, data_dir='./data/', extra_targets=False)
ground_truth = hafner_ground_truth()
num_inducing = 12
use_natural = False
num_times = dataset[0][0].shape[0]
num_replicates = 1
num_genes = len(dataset.gene_names)
num_tfs = 1
t_inducing = torch.linspace(0, 12, num_times, dtype=torch.float64)
t_observed = torch.linspace(0, 12, num_times)
t_predict = torch.linspace(0, 15, 80, dtype=torch.float32)

config = VariationalConfiguration(
    num_samples=70,
    initial_conditions=False
)

inducing_points = torch.linspace(0, 12, num_inducing).repeat(num_tfs, 1).view(num_tfs, num_inducing, 1)
step_size = 5e-1
num_training = dataset.m_observed.shape[-1]
gp_model = generate_multioutput_gp(num_tfs, inducing_points,
                                   zero_mean=False,
                                   initial_lengthscale=2,
                                   gp_kwargs=dict(natural=use_natural))
lfm = TranscriptionLFM(num_genes, gp_model, config, num_training_points=num_training)
plotter = Plotter1d(lfm, dataset.gene_names, style='seaborn')

track_parameters = [
    'raw_basal',
    'raw_decay',
    'raw_sensitivity',
    'gp_model.covar_module.raw_lengthscale',
]
if use_natural:
    variational_optimizer = NGD(lfm.variational_parameters(), num_data=num_training, lr=0.1)
    parameter_optimizer = Adam(lfm.nonvariational_parameters(), lr=0.02)
    optimizers = [variational_optimizer, parameter_optimizer]
else:
    optimizers = [Adam(lfm.parameters(), lr=0.05)]


trainer = VariationalTrainer(lfm, optimizers, dataset, track_parameters=track_parameters)


lfm.train()
# trainer = Trainer(optimizer)
output = trainer.train(500, step_size=5e-1, report_interval=10)

tight_kwargs = dict(bbox_inches='tight', pad_inches=0)
t_predict = torch.linspace(0, 15, 80, dtype=torch.float32)

lfm.eval()
q_m = lfm.predict_m(t_predict, step_size=1e-1)
q_f = lfm.predict_f(t_predict)

plotter.plot_losses(trainer, last_x=200)
nrows = 3
ncols = 3
fig, axes = plt.subplots(nrows=nrows, ncols=ncols, figsize=(10, 6))

row = col = 0
for i in range(8):
    if i == (row+1) * 3:
        row += 1
        col = 0
    ax = axes[row, col]
    plotter.plot_gp(q_m, t_predict, replicate=0, ax=ax,# ylim=(-2, 25.2),
                    color=Colours.line_color, shade_color=Colours.shade_color,
                    t_scatter=dataset.t_observed, y_scatter=dataset.m_observed,
                    only_plot_index=i, num_samples=0)
    col += 1
    ax.set_title(dataset.gene_names[i])
plotter.plot_gp(q_f, t_predict, ax=axes[nrows-1, ncols-1],
                # ylim=(-1, 5),
                transform=softplus,
                num_samples=3,
                color=Colours.line2_color,
                shade_color=Colours.shade2_color)
axes[nrows-1, ncols-1].set_title('Latent force (p53)')
for col in range(ncols):
    axes[nrows-1, col].set_xlabel('Time (h)')

plt.tight_layout()
plt.savefig('./experiments/hafner-multi.pdf', **tight_kwargs)


fig, axes = plt.subplots(nrows=2, ncols=4, figsize=(9, 2.9),
                         gridspec_kw=dict(width_ratios=[1, 1, 0.5, 1.9], wspace=0, hspace=0.75))
row = col = 0
plots = [2, 4, 7, 6]
lbs = [2, 0, 2, 5]
for i in range(4):
    if i == (row+1) * 2:
        row += 1
        col = 0
    ax = axes[row, col]
    plotter.plot_gp(q_m, t_predict, replicate=0, ax=ax,# ylim=(-2, 25.2),
                    color=Colours.line_color, shade_color=Colours.shade_color,
                    t_scatter=dataset.t_observed, y_scatter=dataset.m_observed,
                    only_plot_index=plots[i], num_samples=0)
    ax.set_title(dataset.gene_names[plots[i]])
    mean = q_m.mean.detach().transpose(0, 1)[plots[i]]
    lb = torch.floor(mean.min()-1)
    ub = torch.ceil(mean.max() + 1)
    brange = ub - lb
    ub = int(lb + brange*1.1)
    lb = lbs[i]
    ax.set_ylim(lb, ub)
    ax.set_xlim(0, 15)
    ax.set_yticks([lb, ub])
    ax.figure.subplotpars.wspace = 0
    if col > 0:
        ax.set_yticks([])
        ax.set_xticks([5, 10, 15])

    col += 1
y = [1.5, 4.8, 13.7, 5, 2, 1.4, 3.2, 4, 1.4, 1.5]
# y = y/np.mean(y)*np.mean(p) * 1.75-0.16
# y = scaler.fit_transform(np.expand_dims(y, 0))
axes[0, 3].plot(np.linspace(0, 10, len(y)), y, color=Colours.line2_color)
axes[0, 3].set_ylim(0, 15)
axes[0, 3].set_ylabel('Fold change')
axes[0, 3].set_yticks([0.0, 15.0])
axes[0, 3].set_title('Western blot (Hafner et al., 2017)')
t_temp = torch.linspace(0, 10, 80, dtype=torch.float32)
q_f = lfm.predict_f(t_temp)
plotter.plot_gp(q_f, t_temp, ax=axes[1, 3],
                transform=softplus,
                num_samples=3,
                color=Colours.line2_color,
                shade_color=Colours.shade2_color)
axes[1, 3].set_yticks([0, 4])
for i in range(2):
    axes[i, 3].set_xlim(0, 10)
    axes[i, 2].set_visible(False)

axes[1, 3].set_title('Latent force inference (ours)')
axes[1, 3].set_xlabel('Time (h)')
axes[1, 3].set_ylabel('FPKM $\ell_2$')
axes[1, 3].set_yticks([0.0, 6.0])
axes[1, 3].set_ylim(0.0, 6)
# axes[1, 3].yaxis.set_major_formatter(FormatStrFormatter('  %.0f'))

fig.tight_layout()

plt.savefig('./experiments/hafner-combined.pdf', **tight_kwargs)

labels = ['Basal rates', 'Sensitivities', 'Decay rates']
keys = ['raw_basal', 'raw_sensitivity', 'raw_decay']
constraints = [lfm.positivity, lfm.positivity, lfm.positivity]
kinetics = list()
for i, key in enumerate(keys):
    kinetics.append(
        constraints[i].transform(trainer.parameter_trace[key][-1].squeeze()).numpy())

kinetics = np.array(kinetics)
plotter.plot_double_bar(kinetics, titles=labels,
                        figsize=(9, 3),
                        ground_truths=hafner_ground_truth())
                        # yticks=[
                        #     np.linspace(0, 0.12, 5),
                        #     np.linspace(0, 1.2, 4),
                        #     np.arange(0, 1.1, 0.2),
                        # ])
plt.tight_layout()
# plt.savefig('./hafner-kinetics.pdf', **tight_kwargs)

# plotter.plot_convergence(trainer)


# print(lfm.decay_rate)
print(lfm.sensitivity[6])
for i in range(20):
    plt.plot(lfm.positivity.transform(torch.stack(trainer.parameter_trace['raw_sensitivity'])[:, i]))



from dataclasses import dataclass, field


@dataclass
class Configuration:
    preprocessing_variance: object = None # data processing variances if present (e.g. mmgmos) otherwise None
    latent_data_present:    bool = False  # False for fully inferred latents
    delays:                 bool = False  # True if delay params used
    kernel:                 str = 'rbf'   # Kernel for latent function, rbf/mlp
    initial_conditions:     bool = True   # True if initial conditions for gene mRNA used
    translation:            bool = True   # True if the translation ODE is active
    kinetic_exponential:    bool = False  # True if kinetic params are exponentiated
    kernel_exponential:     bool = False  # True if kernel params are exponentiated


@dataclass
class VariationalConfiguration(Configuration):
    num_samples:            int = 20       # number of samples from the variational distribution


@dataclass
class MCMCConfiguration(Configuration):
    initial_step_sizes: dict = field(default_factory=dict)
    weights: bool = True  # True if weights used



from torch.utils.data import Dataset


class LFMDataset(Dataset):

    def __getitem__(self, index):
        return self.data[index]

    @property
    def num_outputs(self):
        """The number of LFM outputs."""
        return self._num_outputs

    @num_outputs.setter
    def num_outputs(self, value):
        self._num_outputs = value

    @property
    def data(self):
        """
        List of data points, each a tuple(a, b).
        For time-series, a and b are 1-D.
        For spatiotemporal series, a is (2, T) corresponding to a row for time and space, and b is 1-D.
        """
        return self._data

    @data.setter
    def data(self, value):
        self._data = value

    def __len__(self):
        return len(self.data)



from dataclasses import dataclass, field
from matplotlib import pyplot as plt
from matplotlib.animation import FuncAnimation

import numpy as np
import arviz
import networkx as nx
import tensorflow as tf
import random
import seaborn as sns

from alfi.datasets import scaled_barenco_data


@dataclass
class PlotOptions:
    num_plot_genes: int = 20
    num_plot_tfs:   int = 20
    gene_names:     np.array = None
    tf_names:       np.array = None
    num_kinetic_avg:int = 50
    num_hpd:        int = 100
    true_label:     str = 'True'
    tf_present:     bool = True
    for_report:     bool = True
    ylabel:         str = ''
    protein_present:bool = True
    model_label:    str = 'Model'
    kernel_names:   list = field(default_factory=lambda: ['Param 1', 'Param 2'])


class Plotter:
    def __init__(self, data, options: PlotOptions, style='seaborn'):
        self.opt = options
        if self.opt.gene_names is None:
            self.opt.gene_names = np.array([f'Gene {j}' for j in range(data.m_obs.shape[1])])
        if self.opt.tf_names is None:
            self.opt.tf_names = np.array([f'TF {i}' for i in range(data.f_obs.shape[1])])
        self.num_tfs = data.f_obs.shape[1]
        self.num_outputs = data.m_obs.shape[1]
        self.data = data
        self.t_discretised = data.t_discretised.numpy()
        self.t = data.t_observed
        self.common_ind = data.common_indices.numpy()
        palette = sns.color_palette('colorblind')
        self.shade_color = palette[0]
        self.line_color = palette[0]
        self.scatter_color = palette[3]
        self.bar1_color = palette[3]
        self.bar2_color = palette[2]
        plt.style.use(style)
        sns.set(font="CMU Serif")
        plt.rcParams['font.family'] = 'serif'
        plt.rcParams['font.serif'] = 'CMU Serif'

    def plot_kinetics(self, results, kinetic_params, true_k=None, true_hpds=None, title='', xlabels=None, transform=None):
        """
        Parameters:
            results: the results dictionary from the sampler
            kinetic_params: list of strings. The keys in `results` which are kinetic parameters of shape
                            (num_samples, num_outputs, 1)
        """
        if transform is None:
            transform = lambda x: x
        kinetics = transform(np.stack([results[k] for k in kinetic_params]).transpose((1, 2, 0, 3)).squeeze(-1))
        mean_kinetics = np.mean(kinetics[-self.opt.num_kinetic_avg:], axis=0)

        plot_labels = ['Initial Conditions', 'Basal rates', 'Decay rates', 'Sensitivities']
        num_x = mean_kinetics.shape[0]
        width = num_x
        height = 10
        if num_x < 2:
            width = 4
            height = 3
        plt.figure(figsize=(width, height))
        if not self.opt.for_report:
            plt.suptitle(title)

        if mean_kinetics.shape[1] < 4:
            plot_labels = plot_labels[1:]

        hpds = self.plot_bar_hpd(
            kinetics, mean_kinetics,
            xlabels,
            true_var=true_k, true_hpds=true_hpds,
            width=0.2, rotation=60)
        plt.tight_layout(h_pad=2.0)

        return mean_kinetics, hpds

    def plot_bar_hpd(self, var_samples, var, labels, true_var=None, width=0.1, titles=None, 
                     rotation=0, true_hpds=None):
        hpds = list()
        num = var.shape[0]
        plotnum = var.shape[1]*100 + 21
        for i in range(num):
            hpds.append(arviz.hdi(var_samples[-self.opt.num_hpd:, i,:], 0.95))
        hpds = np.array(hpds)
        hpds = abs(hpds - np.expand_dims(var, 2))
        for k in range(var_samples.shape[2]):
            plt.subplot(plotnum)
            plotnum += 1
            plt.bar(np.arange(num)-width, var[:, k], width=2*width, tick_label=labels, color=self.bar1_color, label=self.opt.model_label)
            plt.errorbar(np.arange(num)-width, var[:, k], hpds[:, k].swapaxes(0,1), fmt='none', capsize=5, color='black')
            plt.xlim(-1, num)
            plt.xticks(rotation=rotation)
            if titles is not None:
                plt.title(titles[k])
            if true_var is not None:
                plt.bar(np.arange(num)+width, true_var[:, k], width=2*width, color=self.bar2_color, align='center', label=self.opt.true_label)
                if true_hpds is not None:
                    plt.errorbar(np.arange(num)+width, true_var[:, k], true_hpds[:, k].swapaxes(0,1), fmt='none', capsize=5, color='black')
                plt.legend()
        plt.tight_layout()
        return hpds

    def plot_convergence(self, results, param_names, title='', transform=None):
        if transform is None:
            transform = lambda x: x
        params = [transform(results[param]) for param in param_names]
        num = len(params)
        plt.figure(figsize=(8, 3 * num))
        plt.suptitle(title)

        for j in range(num):
            param = params[j]
            ax = plt.subplot(num, 1, j + 1)
            ax.set_title(param_names[j])
            plt.plot(np.squeeze(param[:, :]))  # , label=labels[v])
            plt.legend()
        plt.tight_layout()

    def plot_samples(self, samples, titles, num_samples, scatters=None,
                     scatter_args=None, legend=True, margined=False, sample_gap=2):
        if scatter_args is None:
            scatter_args = dict()
        num_components = samples[0].shape[0]
        subplot_shape = ((num_components+2)//3, 3)
        if self.opt.for_report:
            subplot_shape = ((num_components+1)//2, 2)
        if num_components <= 1:
            subplot_shape = (1, 1) 
        for j in range(num_components):
            ax = plt.subplot(subplot_shape[0], subplot_shape[1], 1+j)
            plt.title(titles[j])
            if scatters is not None:
                plt.scatter(self.t_discretised[self.common_ind], scatters[j], marker='x', label='Observed', color=self.scatter_color, **scatter_args)
            # plt.errorbar([n*10+n for n in range(7)], Y[j], 2*np.sqrt(Y_var[j]), fmt='none', capsize=5)

            for s in range(1, sample_gap*num_samples, sample_gap):
                kwargs = {}
                if s == 1:
                    kwargs = {'label':'Samples'}

                plt.plot(self.t_discretised, samples[-s, j, :], color=self.line_color, alpha=0.5, **kwargs)
            if j % subplot_shape[1] == 0:
                plt.ylabel(self.opt.ylabel)

            # HPD:
            bounds = arviz.hdi(np.expand_dims(samples[-self.opt.num_hpd:,j,:], 0), 0.95)
            plt.fill_between(self.t_discretised, bounds[:, 0], bounds[:, 1], color='grey', alpha=0.3, label='95% credibility interval')

            plt.xticks(self.t)
            ax.set_xticklabels(self.t)
            if margined:
                plt.ylim(min(samples[-1, j])-2, max(samples[-1, j]) + 2)
            # if self.opt.for_report:
                # plt.ylim(-0.2, max(samples[-1, j]) + 0.2)
            plt.xlabel('Time (h)')
            if legend:
                plt.legend()
        plt.tight_layout()

    def plot_outputs(self, m_preds, replicate=0, height_mul=3, width_mul=2, indices=None):
        m_preds = m_preds[:, replicate]
        scatters = self.data.m_obs[replicate]
        if indices:
            m_preds = m_preds[:, indices]
            scatters = scatters[indices]
        height = np.ceil(m_preds.shape[1]/3)
        width = 4 if self.opt.for_report else 7
        plt.figure(figsize=(width*width_mul, height*height_mul))
        if not self.opt.for_report:
            plt.suptitle('Genes')
        names = self.opt.gene_names[indices] if indices is not None else self.opt.gene_names
        self.plot_samples(m_preds, names, self.opt.num_plot_genes, 
                          scatters=scatters, legend=not self.opt.for_report)

    def plot_latents(self, f_samples, replicate=0, plot_barenco=False, sample_gap=2):
        f_samples = f_samples[:, replicate]
        num_tfs = self.data.f_obs.shape[1]
        width = 6*min(num_tfs, 3)
        figsize=(width, 4*np.ceil(num_tfs/3))
        if self.opt.for_report and num_tfs == 1:
            figsize=(6, 4)
        plt.figure(figsize=figsize)
        scatter_args = {'s': 60, 'linewidth': 2}
        self.plot_samples(f_samples, self.opt.tf_names, self.opt.num_plot_tfs,
                          scatters=self.data.f_obs[replicate] if self.opt.tf_present else None,
                          scatter_args=scatter_args, margined=True, sample_gap=sample_gap)

        # if 'σ2_f' in model.params._fields:
        #     σ2_f = model.params.σ2_f.value
        #     plt.errorbar(t_discretised[common_indices], f_observed[0], 2*np.sqrt(σ2_f[0]),
        #                 fmt='none', capsize=5, color='blue')
        # else:
        #     σ2_f = σ2_f_pre
        # for i in range(num_tfs):
        #     f_obs = self.data.f_obs[replicate, i]
        #     if scale_observed:
        #         f_obs = f_obs / np.mean(f_obs) * np.mean(f_i)

        if plot_barenco:
            barenco_f, _ = scaled_barenco_data(np.mean(f_samples[-10:], axis=0))
            plt.scatter(self.t_discretised[self.common_ind], barenco_f, marker='x', s=60, linewidth=3, label='Barenco et al.')

        plt.tight_layout()

    def moving_average(self, a, n=3) :
        ret = np.cumsum(a, dtype=float)
        ret[n:] = ret[n:] - ret[:-n]
        return ret[n - 1:] / n

    def plot_noises(self, σ2_m_samples, σ2_f_samples):
        plt.figure(figsize=(5, 3))
        for j in range(self.opt.gene_names.shape[0]):
            ma = self.moving_average(σ2_m_samples[-1000:, j], n=20)
            plt.plot(ma, label=self.opt.gene_names[j])
        if self.opt.gene_names.shape[0] < 8:
            plt.legend()
        if σ2_f_samples is not None:
            plt.figure(figsize=(5, 3))
            for j in range(σ2_f_samples.shape[1]):
                plt.plot(σ2_f_samples[:, j])
            plt.legend()

    def plot_weights(self, weights):
        plt.figure()
        w = weights[0]
        w_0 = weights[1]
        for j in range(self.opt.gene_names.shape[0]):
            plt.plot(w[:, j], label=self.opt.gene_names[j])
        if self.opt.gene_names.shape[0] < 8:
            plt.legend()
        plt.title('Interaction weights')
        plt.figure()
        for j in range(self.opt.gene_names.shape[0]):
            plt.plot(w_0[:,j])
        plt.title('Interaction bias')

    '''
    Plots convergence with histogram on the side
    Args:
      samples: 1D array
      lims: tuple (lower, upper)
    '''
    def plot_convergence_hist(self, samples, lims=None, fig_height=6, fig_width=8, color='slategrey'):
        left, width = 0.1, 0.65
        bottom, height = 0.1, 0.65
        spacing = 0.005
        rect_scatter = [left, bottom, width, height]
        rect_histy = [left + width + spacing, bottom, 0.2, height]
        plt.figure(figsize=(fig_width, fig_height))

        ax_scatter = plt.axes(rect_scatter)
        ax_scatter.tick_params(direction='in', top=True, right=True)
        ax_histy = plt.axes(rect_histy)
        ax_histy.tick_params(direction='in', labelleft=False)

        # the scatter plot:
        ax_scatter.plot(samples, color=color)

        # now determine nice limits by hand:
        binwidth = 0.25
        if lims is None:
            lims = (0, np.ceil(samples.max() / binwidth) * binwidth)
        # ax_scatter.set_xlim((-lim, lim))
        ax_scatter.set_ylim(lims)
        ax_histy.hist(samples, orientation='horizontal', color=color, bins='auto')#, bins=bins)

        ax_histy.set_ylim(ax_scatter.get_ylim())

    def anim_latent(self, samples, index=0, replicate=0, interval=10):
        fig = plt.figure()
        s_min, s_max = np.min(samples[:, 0, 0, :]), np.max(samples[:, replicate, index, :])
        ax = plt.axes(xlim=(-1, 13), ylim=(s_min-0.4, s_max+0.4))
        line, = ax.plot([], [], lw=3)
        def init():
            line.set_data([], [])
            return line,
        def animate(i):
            x = np.linspace(0, 12, samples.shape[3])
            y = samples[i*interval, replicate, index]
            line.set_data(x, y)
            return line,
        anim = FuncAnimation(fig, animate, init_func=init,
                                    frames=samples.shape[0]//interval, interval=50, blit=True)
        return anim.to_html5_video()

    def plot_grn(self, results, kinetic_params, use_basal=True, use_sensitivities=True):
        G = nx.DiGraph()
        random.seed(42)
        np.random.seed(42)
        nodes, node_colors, sizes, edges, colors = list(), list(), list(), list(), list()

        for i in range(self.opt.tf_names.shape[0]):
            nodes.append(self.opt.tf_names[i])
            node_colors.append('slategrey')
            sizes.append(1000)

        for j in range(self.opt.gene_names.shape[0]):
            nodes.append(self.opt.gene_names[j])
            node_colors.append('chocolate')

        kinetics = np.stack([results[k] for k in kinetic_params]).transpose((1, 2, 0, 3)).squeeze(-1)
        mean_kinetics = np.mean(kinetics[-self.opt.num_kinetic_avg:], axis=0)

        b = mean_kinetics[:, 0]
        s = mean_kinetics[:, 2]
        if not use_sensitivities:
            w = np.mean(results.weights[0][-100:], axis=0)
            w_min = tf.math.reduce_min(w).numpy()
            w_diff = tf.math.reduce_max(w).numpy() - w_min

        s_min = np.min(s)
        s_diff = max(s) - s_min
        b_min = np.min(b)
        b_diff = max(b) - b_min

        for j in range(self.num_outputs):
            for i in range(self.num_tfs):
                edge = (self.opt.tf_names[i], self.opt.gene_names[j])
                weight = (s[j]-s_min) / s_diff if use_sensitivities else (w[j, i]-w_min) / w_diff
                colors.append(f'{1-weight}')
                edges.append(edge)
            if use_basal:
                sizes.append(int(700 + 1700 * (b[j]-b_min)/b_diff))
        G.add_nodes_from(nodes)
        G.add_edges_from(edges)
        node_size = sizes if use_basal else 1000
        pos = nx.spring_layout(G, seed=42)
        # pos = nx.nx_pydot.graphviz_layout(G, prog='dot')
        nx.draw(G, pos=pos, edge_color=colors, node_color=node_colors, node_size=node_size, with_labels=True)



import torch
from gpytorch.optim import NGD
from torch.optim import Adam
from matplotlib import pyplot as plt
from gpytorch.mlls.exact_marginal_log_likelihood import ExactMarginalLogLikelihood
from torch.nn.functional import l1_loss

from alfi.datasets import P53Data
from alfi.configuration import VariationalConfiguration
from alfi.models import generate_multioutput_gp
from alfi.plot import Plotter1d, Colours, tight_kwargs
from alfi.trainers import VariationalTrainer, PreEstimator
from alfi.models import ExactLFM
from alfi.trainers import ExactTrainer
from alfi.impl.odes import TranscriptionLFM


plt.rcParams.update({'font.size': 82})
plt.rc('font', size=15)          # controls default text sizes
plt.rc('axes', titlesize=15)     # fontsize of the axes title
plt.rc('axes', labelsize=15)    # fontsize of the x and y labels
plt.rc('xtick', labelsize=14)    # fontsize of the tick labels
plt.rc('ytick', labelsize=14)    # fontsize of the tick labels
plt.rc('legend', fontsize=20)    # legend fontsize
plt.rc('figure', titlesize=20)  # fontsize of the figure title


class ConstrainedTrainer(VariationalTrainer):
    def after_epoch(self):
        with torch.no_grad():
            sens = torch.tensor(1.)
            dec = torch.tensor(0.8)
            self.lfm.raw_sensitivity[3] = self.lfm.positivity.inverse_transform(sens)
            self.lfm.raw_decay[3] = self.lfm.positivity.inverse_transform(dec)
        super().after_epoch()


dataset = P53Data(replicate=0, data_dir='./data')
num_inducing = 20

num_replicates = 1
num_genes = len(dataset.gene_names)
num_tfs = 1
num_times = dataset[0][0].shape[0]
t_end = dataset.t_observed[-1]
use_natural = True

config = VariationalConfiguration(
    # preprocessing_variance=dataset.variance,
    num_samples=80,
    initial_conditions=False
)

inducing_points = torch.linspace(0, t_end, num_inducing).repeat(num_tfs, 1).view(num_tfs, num_inducing, 1)
t_predict = torch.linspace(0, t_end+3, 80, dtype=torch.float32)
step_size = 5e-1
num_training = dataset.m_observed.shape[-1]
gp_model = generate_multioutput_gp(num_tfs, inducing_points, gp_kwargs=dict(natural=use_natural))

lfm = TranscriptionLFM(num_genes, gp_model, config,
                       initial_basal=0.1,
                       initial_decay=0.1,
                       initial_sensitivity=2,
                       num_training_points=num_training)

plotter = Plotter1d(lfm, dataset.gene_names, style='seaborn')

track_parameters = [
    'raw_basal',
    'raw_decay',
    'raw_sensitivity',
    'gp_model.covar_module.raw_lengthscale',
]
if use_natural:
    variational_optimizer = NGD(lfm.variational_parameters(), num_data=num_training, lr=0.09)
    parameter_optimizer = Adam(lfm.nonvariational_parameters(), lr=0.02)
    optimizers = [variational_optimizer, parameter_optimizer]
    pre_variational_optimizer = NGD(lfm.variational_parameters(), num_data=num_training, lr=0.1)
    pre_parameter_optimizer = Adam(lfm.nonvariational_parameters(), lr=0.005)
    pre_optimizers = [pre_variational_optimizer, pre_parameter_optimizer]

else:
    optimizers = [Adam(lfm.parameters(), lr=0.05)]
    pre_optimizers = [Adam(lfm.parameters(), lr=0.05)]

trainer = ConstrainedTrainer(lfm, optimizers, dataset, track_parameters=track_parameters)
pre_estimator = PreEstimator(lfm, pre_optimizers, dataset, track_parameters=track_parameters)

lfm.loss_fn.num_data = num_training
trainer.train(200, report_interval=10, step_size=step_size);

q_m = lfm.predict_m(t_predict, step_size=1e-1)
q_f = lfm.predict_f(t_predict)


exact_lfm = ExactLFM(dataset, dataset.variance.reshape(-1))
optimizer = torch.optim.Adam(exact_lfm.parameters(), lr=0.07)

loss_fn = ExactMarginalLogLikelihood(exact_lfm.likelihood, exact_lfm)

track_parameters = [
    'mean_module.raw_basal',
    'covar_module.raw_decay',
    'covar_module.raw_sensitivity',
    'covar_module.raw_lengthscale',
]
exact_trainer = ExactTrainer(exact_lfm, [optimizer], dataset, loss_fn=loss_fn, track_parameters=track_parameters)
exact_plotter = Plotter1d(exact_lfm, dataset.gene_names)


exact_lfm.train()
exact_lfm.likelihood.train()
exact_trainer.train(epochs=150, report_interval=10);

exact_q_m = exact_lfm.predict_m(t_predict)
exact_q_f = exact_lfm.predict_f(t_predict)

fig, axes = plt.subplots(nrows=2, ncols=4, figsize=(9, 2.9),
                         gridspec_kw=dict(width_ratios=[1, 1, 0.3, 1.8], wspace=0, hspace=0.75))

row = 0
col = 0
ub = [3.5] * 4
for i in range(4):
    if i == 2:
        row += 1
        col = 0
    ax = axes[row, col]
    plotter.plot_gp(q_m, t_predict, replicate=0, ax=ax,
                    color=Colours.line_color, shade_color=Colours.shade_color,
                    t_scatter=dataset.t_observed, y_scatter=dataset.m_observed,
                    num_samples=0, only_plot_index=i)
    ax.set_ylim([-0.2, ub[i]])
    ax.set_yticks([0, 3])
    ax.set_title(dataset.gene_names[i])
    ax.set_xlim(-0.4, 15)
    if col > 0:
        ax.set_yticks([])
        ax.set_xticks([5, 10, 15])
    else:
        ax.set_xticks([0, 5, 10, 15])
    col += 1
plotter.plot_gp(q_f, t_predict, ax=axes[1, 3],
                ylim=(-1, 3.2),
                num_samples=0,
                color=Colours.line2_color,
                shade_color=Colours.shade2_color)
plotter.plot_gp(exact_q_f, t_predict, ax=axes[0, 3],
                ylim=(-1, 3.2), color=Colours.line2_color,
                shade_color=Colours.shade2_color)
titles = ['(Lawrence et al., 2007)', 'inference (ours)']
for i in range(2):
    axes[i, 3].set_title(f'Latent force {titles[i]}')
    axes[i, 3].set_yticks([-1, 3])
    axes[i, 3].set_xlim(0, 15)
    axes[i, 3].set_xticks([0, 5, 10, 15])
    axes[i, 2].set_visible(False)
axes[1, 3].set_xlabel('Time (h)')

plt.savefig('./experiments/barenco-combined.pdf', **tight_kwargs)

print(l1_loss(q_f.mean, exact_q_f.mean).item())



import gpytorch

from .torte import Trainer as TorteTrainer


class Trainer(TorteTrainer):
    def print_extra(self):
        if isinstance(self.model, gpytorch.models.GP):
            kernel = self.model.covar_module
            print(f'λ: {str(kernel.lengthscale.view(-1).detach().numpy())}', end='')
        elif hasattr(self.model, 'gp_model'):
            print(f'kernel: {self.model.summarise_gp_hyp()}', end='')
        super().print_extra()



#%%
from experiments.partial import build_partial, plot_partial, pretrain_partial

from alfi.datasets import DrosophilaSpatialTranscriptomics, HomogeneousReactionDiffusion
from alfi.models import TrainMode
from matplotlib import pyplot as plt
import torch
from alfi.configuration import VariationalConfiguration

mrna_q2s = list()
mrna_cias = list()
protein_q2s = list()
protein_cias = list()

gene = 'kni'

dataset = DrosophilaSpatialTranscriptomics(
    gene=gene, data_dir='./data', scale=True)
disc = dataset.disc

params = dict(lengthscale=10,
              **params[gene],
              parameter_grad=False,
              warm_epochs=-1,
              natural=True,
              zero_mean=True,
              clamp=True)

lfm, trainer, plotter = build_partial(
    dataset,
    params)

tx = trainer.tx
orig_data = dataset.orig_data.squeeze().t()
num_t = tx[0, :].unique().shape[0]
num_x = tx[1, :].unique().shape[0]

num_t_orig = orig_data[:, 0].unique().shape[0]
num_x_orig = orig_data[:, 1].unique().shape[0]

y_target = trainer.y_target[0]
y_matrix = y_target.view(num_t_orig, num_x_orig)

for i in range(1):
    lfm, trainer, plotter = build_partial(
        dataset,
        params)


    ts = tx[0, :].unique().numpy()
    xs = tx[1, :].unique().numpy()
    extent = [ts[0], ts[-1], xs[0], xs[-1]]

    times = pretrain_partial(dataset, lfm, trainer, params)

    trainer.plot_outputs = False
    lfm.set_mode(TrainMode.NORMAL)
    trainer.train(150, report_interval=10)
    lfm.save(f'./{gene}{i}')
    # from alfi.plot import tight_kwargs
    # plot_partial(dataset, lfm, trainer, plotter, Path('./'), params)
    #
    # plt.savefig(Path('./') / f'kinetics-{gene}-{i}.pdf', **tight_kwargs)

    # from alfi.utilities.torch import q2, cia
    # f = lfm(tx)
    # f_mean = f.mean.detach()
    # f_var = f.variance.detach()
    # y_target = trainer.y_target[0]

    print('Run ', i)
    # protein_q2 = q2(y_target.squeeze(), f_mean.squeeze())
    protein_q2 = trainer.prot_q2_best
    # protein_cia = cia(y_target.squeeze(), f_mean.squeeze(), f_var.squeeze())
    protein_cia = trainer.cia[1]
    print('Protein Q2', protein_q2)
    print('Protein CA', protein_cia)
    protein_q2s.append(protein_q2.item())
    protein_cias.append(protein_cia.item())

    # gp = lfm.gp_model(tx.t())
    # lf_target = orig_data[trainer.t_sorted, 2]
    # f_mean = gp.mean.detach()
    # f_var = gp.variance.detach()

    # mrna_q2 = q2(lf_target.squeeze(), f_mean.squeeze())
    mrna_q2 = trainer.mrna_q2_best
    # mrna_cia = cia(lf_target.squeeze(), f_mean.squeeze(), f_var.squeeze())
    mrna_cia = trainer.cia[0]
    print('mRNA Q2', mrna_q2)
    print('mRNA CA', mrna_cia)
    mrna_q2s.append(mrna_q2.item())
    mrna_cias.append(mrna_cia.item())

protein_q2s = torch.tensor(protein_q2s)
protein_cias = torch.tensor(protein_cias)
mrna_q2s = torch.tensor(mrna_q2s)
mrna_cias = torch.tensor(mrna_cias)
print('Final')

print(f'Protein Q2: {protein_q2s.mean(0):.04f} pm {protein_q2s.std(0):.04f}')
print(f'Protein CIA: {protein_cias.mean(0):.04f} pm {protein_cias.std(0):.04f}')
print(f'mRNA Q2: {mrna_q2s.mean(0):.04f} pm {mrna_q2s.std(0):.04f}')
print(f'mRNA CIA: {mrna_cias.mean(0):.04f} pm {mrna_cias.std(0):.04f}')

torch.save(protein_q2s, f'experiments/{gene}_protein_q2')
torch.save(protein_cias, f'experiments/{gene}_protein_cia')
torch.save(mrna_q2s, f'experiments/{gene}_mrna_q2')
torch.save(mrna_cias, f'experiments/{gene}_mrna_cia')



import tensorflow as tf
from tensorflow import math as tfm
from tensorflow_probability import distributions as tfd

from alfi.datasets import DataHolder
from alfi.utilities.tf import rotate, logit, logistic, LogisticNormal, inverse_positivity, \
    save_object
from alfi.configuration import MCMCConfiguration
from alfi.mcmc.parameter import Parameter
from alfi.mcmc.samplers import HMCSampler, LatentGPSampler, DelaySampler, GibbsSampler
from alfi.mcmc.samplers.mixed import MixedSampler
from alfi.mcmc.gp.gp_kernels import GPKernelSelector
from alfi.mcmc.models import MCMCLFM

import numpy as np
import pickle

PI = tf.constant(np.pi, dtype='float64')
f64 = np.float64


class TranscriptionRegulationLFM(MCMCLFM):
    """
    An updated version of the Metropolis-Hastings model from Titsias et al. (2012) using a mixed sampler
    """
    def __init__(self, data: DataHolder, options: MCMCConfiguration):
        super().__init__(data, options)
        self.N_p = data.t_discretised.shape[0]
        step_sizes = self.options.initial_step_sizes
        logistic_step_size = step_sizes['nuts'] if 'nuts' in step_sizes else 0.00001
        self.subsamplers = list()
        self.dtype = 'float64'
        # Kinetics
        # if options.kinetic_exponential:
        #     kinetic_transform = lambda x: tf.exp(logit(x))
        # else:
        kinetic_transform = logit

        basal_rate = Parameter(
            'basal',
            LogisticNormal(0.01, 30),
            tf.random.uniform((self.num_outputs, 1), 0.75, 0.85, dtype=self.dtype),
            transform=kinetic_transform
        )
        sensitivity = Parameter(
            'sensitivity',
            LogisticNormal(0.01, 30),
            tf.random.uniform((self.num_outputs, 1), 0.75, 0.85, dtype=self.dtype),
            transform=kinetic_transform
        )
        decay_rate = Parameter(
            'decay',
            LogisticNormal(0.01, 30),
            tf.random.uniform((self.num_outputs, 1), 0.75, 0.85, dtype=self.dtype),
            transform=kinetic_transform
        )
        kinetics = [basal_rate, decay_rate, sensitivity]

        # Add optional kinetic parameters
        if options.initial_conditions:
            self.initial_conditions = Parameter(
                'initial',
                LogisticNormal(0.01, 30),
                tf.random.uniform((self.num_outputs, 1), 0.75, 0.85, dtype=self.dtype),
                transform=kinetic_transform
            )
            kinetics.append(self.initial_conditions)
        if options.translation:
            self.protein_decay = Parameter(
                'protein_decay',
                LogisticNormal(0.1, 7),
                0.8 * tf.ones((self.num_tfs, 1), dtype=self.dtype),
                transform=logit
            )
            kinetics.append(self.protein_decay)

        kinetics_subsampler = HMCSampler(self.likelihood, kinetics, logistic_step_size)
        self.subsamplers.append(kinetics_subsampler)

        # Weights
        if options.weights:
            self.weights = Parameter(
                'w',
                LogisticNormal(f64(-2), f64(2)),
                logistic(1 * tf.ones((self.num_outputs, self.num_tfs), dtype=self.dtype))
            )
            self.weights_biases = Parameter(
                'w_0',
                LogisticNormal(f64(-0.8), f64(0.8)),
                logistic(0 * tf.ones(self.num_outputs, dtype=self.dtype))
            )
            weights = [self.weights, self.weights_biases]
            weights_subsampler = HMCSampler(self.likelihood, weights, logistic_step_size)
            self.subsamplers.append(weights_subsampler)

        # Latent function & GP hyperparameters
        self.kernel_selector = GPKernelSelector(data, options)
        kernel_initial = self.kernel_selector.initial_params()

        f_step_size = step_sizes['latents'] if 'latents' in step_sizes else 20
        latent_likelihood = self.tfs_likelihood if options.latent_data_present else None
        latents_initial = 0.3 * tf.ones((self.num_replicates, self.num_tfs, self.N_p), dtype=self.dtype)

        latents_initial = [latents_initial, *kernel_initial]

        latents = Parameter('latent', None, latents_initial)
        latents_sampler = LatentGPSampler(self.likelihood, latent_likelihood,
                                          latents, self.kernel_selector,
                                          f_step_size*tf.ones(self.N_p, dtype=self.dtype),
                                          kernel_exponential=options.kernel_exponential)

        self.subsamplers.append(latents_sampler)

        if options.delays:
            delay_prior = tfd.InverseGamma(f64(0.01), f64(0.01)) #TODO choose between these two
            delay_prior = tfd.Exponential(f64(0.3))
            delay = Parameter('delay', delay_prior,
                              0.6 * tf.ones(self.num_tfs, dtype=self.dtype))
            delay_sampler = DelaySampler(self.likelihood, delay, 0, 10)
            self.subsamplers.append(delay_sampler)
        σ2_f = None
        if not options.preprocessing_variance:
            def f_sq_diff_fn():
                f_pred = inverse_positivity(self.parameter_state['latent'][0])
                sq_diff = tfm.square(self.data.f_obs - tf.gather(f_pred, self.data.common_indices, axis=2))
                return tf.reduce_sum(sq_diff, axis=0)
            σ2_f = Parameter('σ2_f',
                             tfd.InverseGamma(f64(0.01), f64(0.01)),
                             1e-4*tf.ones((self.num_tfs,1), dtype=self.dtype))
            σ2_f_sampler = GibbsSampler(σ2_f, f_sq_diff_fn, self.N_p)
            self.subsamplers.append(σ2_f_sampler)
        # White noise for genes
        if not options.preprocessing_variance:
            def m_sq_diff_fn():
                m_pred = self.predict_m(**self.parameter_state)
                sq_diff = tfm.square(self.data.m_obs - tf.gather(m_pred, self.data.common_indices, axis=2))
                return tf.reduce_sum(sq_diff, axis=0)
            σ2_m = Parameter('σ2_m', tfd.InverseGamma(f64(0.01), f64(0.01)),
                             1e-3 * tf.ones((self.num_outputs, 1), dtype=self.dtype))
            σ2_m_sampler = GibbsSampler(σ2_m, m_sq_diff_fn, self.N_p)
        else:
            σ2_m = Parameter('σ2_m', LogisticNormal(f64(1e-5), f64(1)),
                             0.7 * tf.ones((self.num_outputs, 1), dtype=self.dtype),
                             transform=logit)
            σ2_m_sampler = HMCSampler(self.likelihood, [σ2_m], logistic_step_size)

        self.subsamplers.append(σ2_m_sampler)

        def iteration_callback(current_state):
            self.parameter_state = current_state

        self.sampler = MixedSampler(self.subsamplers, iteration_callback=iteration_callback)

    def sample(self, T=2000, **kwargs):
        return self.sampler.sample(T, **kwargs)

    @tf.function
    def calculate_protein(self, fbar, protein_decay, delay):  # Calculate p_i vector
        t_discretised = self.data.t_discretised
        f_i = inverse_positivity(fbar)
        δ_i = tf.reshape(protein_decay, (-1, 1))
        if self.options.delays:
            # Add delay
            delay = tf.cast(delay, 'int32')

            for r in range(self.num_replicates):
                f_ir = rotate(f_i[r], -delay)
                mask = ~tf.sequence_mask(delay, f_i.shape[2])
                f_ir = tf.where(mask, f_ir, 0)
                mask = np.zeros((self.num_replicates, 1, 1), dtype=self.dtype)
                mask[r] = 1
                f_i = (1 - mask) * f_i + mask * f_ir

        # Approximate integral (trapezoid rule)
        resolution = t_discretised[1] - t_discretised[0]
        sum_term = tfm.multiply(tfm.exp(δ_i * t_discretised), f_i)
        cumsum = 0.5 * resolution * tfm.cumsum(sum_term[:, :, :-1] + sum_term[:, :, 1:], axis=2)
        integrals = tf.concat([tf.zeros((self.num_replicates, self.num_tfs, 1), dtype=self.dtype), cumsum], axis=2)
        exp_δt = tfm.exp(-δ_i * t_discretised)
        p_i = exp_δt * integrals
        return p_i

    @tf.function
    def predict_m(self,
                  initial, basal, decay, sensitivity,
                  protein_decay, latent, **optional_parameters):
        # tf.print(initial, basal, decay, sensitivity)
        if self.options.kinetic_exponential:
            initial = tf.exp(initial)
            basal = tf.exp(basal)
            decay = tf.exp(decay)
            sensitivity = tf.exp(sensitivity)
        # tf.print('p', initial, basal, decay, sensitivity)
        t_discretised = self.data.t_discretised
        fbar = latent[0]
        p_i = inverse_positivity(fbar)
        if self.options.translation:
            delay = optional_parameters['delay'] if self.options.delays else None
            p_i = self.calculate_protein(fbar, protein_decay, delay)

        # Calculate m_pred
        resolution = t_discretised[1] - t_discretised[0]
        if self.options.weights:
            w = optional_parameters['w']
            w_0 = optional_parameters['w_0']
            interactions = tf.matmul(w, tfm.log(p_i + 1e-100)) + w_0
            G = tfm.sigmoid(interactions)  # TF Activation Function (sigmoid)
        else:
            G = tf.tile(p_i, (1, self.num_outputs, 1))

        sum_term = G * tfm.exp(decay * t_discretised)
        integrals = tf.concat([tf.zeros((self.num_replicates, self.num_outputs, 1), dtype=self.dtype),  # Trapezoid rule
                               0.5 * resolution * tfm.cumsum(sum_term[:, :, :-1] + sum_term[:, :, 1:], axis=2)], axis=2)
        exp_dt = tfm.exp(-decay * t_discretised)
        integrals = tfm.multiply(exp_dt, integrals)

        m_pred = basal / decay + sensitivity * integrals
        if self.options.initial_conditions:
            m_pred += tfm.multiply((initial - basal / decay), exp_dt)
        return m_pred

    @tf.function
    def _genes(self, σ2_m=None, **parameter_state):
        m_pred = self.predict_m(σ2_m=σ2_m, **parameter_state)
        # tf.print('paramstate', parameter_state)
        sq_diff = tfm.square(self.data.m_obs - tf.gather(m_pred, self.data.common_indices, axis=2))

        variance = tf.reshape(σ2_m, (-1, 1))
        # tf.print(variance)
        if self.preprocessing_variance:
            variance = variance + self.data.σ2_m_pre  # add PUMA variance
        log_lik = -0.5 * tfm.log(2 * PI * variance) - 0.5 * sq_diff / variance
        log_lik = tf.reduce_mean(log_lik)
        return log_lik

    @tf.function  # (experimental_compile=True)
    def likelihood(self, **parameters):
        """
        Likelihood of the form:
        N(m(t), s(t))
        where m(t) = b/d + (a - b/d) exp(-dt) + s int^t_0 G(p(u); w) exp(-d(t-u)) du
        """
        parameter_state = {**self.parameter_state, **parameters}
        return self._genes(**parameter_state)

    @tf.function  # (experimental_compile=True)
    def tfs_likelihood(self, **parameters):
        """
        Computes log-likelihood of the transcription factors.
        """
        parameter_state = {**self.parameter_state, **parameters}
        # tf.print(
        #     'param', parameters['latent'],
        #     'self', self.parameter_state['latent'],
        #     'state', parameter_state['latent']
        #  )
        σ2_f = parameter_state['σ2_f']
        latent = parameter_state['latent']

        # assert self.options.tf_mrna_present
        if not self.preprocessing_variance:
            variance = tf.reshape(σ2_f, (-1, 1))
        else:
            variance = self.data.σ2_f_pre
        f_pred = inverse_positivity(latent[0])
        sq_diff = tfm.square(self.data.f_obs - tf.transpose(tf.gather(tf.transpose(f_pred), self.data.common_indices)))
        log_lik = -0.5 * tfm.log(2 * PI * variance) - 0.5 * sq_diff / variance
        log_lik = tf.reduce_sum(log_lik)

        return log_lik

    def sample_proteins(self, results, num_results):
        p_samples = list()
        for i in range(1, num_results + 1):
            delta = results['delay'][i] if results['delay'] is not None else None
            p_samples.append(self.likelihood.calculate_protein(results.fbar[-i],
                                                               results.k_fbar[-i], delta))
        return np.array(p_samples)

    def sample_latents(self, results, num_results, step=1):
        m_preds = list()
        for i in range(1, num_results, step):
            m_preds.append(self.predict_m_with_results(results, i))
        return np.array(m_preds)

    def results(self, burnin=0):
        results = dict()
        for i, subsampler in enumerate(self.subsamplers):
            group_samples = self.sampler.samples[i]
            for j, param in enumerate(subsampler.param_group):
                samples = group_samples[j]
                if type(samples) is list:
                    samples = [s.numpy()[-burnin:] for s in samples]
                else:
                    samples = samples.numpy()[-burnin:]
                results[param.name] = param.transform(samples)

        return results

    def predict_m_with_results(self, results, i=1):
        parameter_state = self.parameter_state
        for key in results:
            result = results[key]
            if type(result) is list:
                parameter_state[key] = [result[0][-1]]
            else:
                parameter_state[key] = result[-1]
        return self.predict_m(**parameter_state)

    def save(self, name):
        save_object({'samples': self.samples, 'is_accepted': self.is_accepted}, f'custom-{name}')

    @staticmethod
    def load(name, args):
        model = TranscriptionRegulationLFM(*args)

        import os
        path = os.path.join(os.getcwd(), 'saved_models')
        fs = [os.path.join(path, f) for f in os.listdir(path) if f.startswith(f'custom-{name}')]
        files = sorted(fs, key=os.path.getmtime)
        with open(files[-1], 'rb') as f:
            saved_model = pickle.load(f)
            model.samples = saved_model['samples']
            model.is_accepted = saved_model['is_accepted']
        for param in model.active_params:
            index = model.state_indices[param.name]
            param_samples = model.samples[index]
            if type(param_samples) is list:
                param_samples = [[param_samples[i][-1] for i in range(len(param_samples))]]

            param.value = param_samples[-1]

        return model

    @staticmethod
    def initialise_from_state(args, state):
        model = TranscriptionRegulationLFM(*args)
        model.is_accepted = state.is_accepted
        model.samples = state.samples
        return model
